/*------------------------------------------------------*/
/*							*/
/*			CASE MATCHING			*/
/*							*/
/* This file contains the case matching functions.	*/
/* The user enters the values of the case fields to be	*/
/* matched. The computer then attemps to retrieve a case*/
/* from the case base to be used as a ballpark solution.*/
/*							*/
/* Modified for version 16. Two enumeration symbols	*/
/* can only be similar if neither is used as an		*/
/* abstraction symbol.					*/
/*							*/
/* Modified for version 17. The search now uses the	*/
/* index structure to find cases. Fields which are used	*/
/* as indexes must not be left undefined.		*/
/*							*/
/* Modified for version 19. When performing an index	*/
/* search on an abstraction symbol, the specialisations	*/
/* are returned regardless of whether the abstraction	*/
/* was found or not.					*/
/*							*/
/* Modified for version 20. The program should not crash*/
/* out when the user makes a typing error.		*/
/*							*/
/* Modified for version 31. The user now inputs pop-11	*/
/* style lists						*/
/*							*/
/* Modified for version 35. Now checks for newlines in	*/
/* strings within lists.				*/
/*							*/
/* Modified for version 36. Get rid of compilation	*/
/* warning messages. It also abollishes the result	*/
/* value.						*/
/*							*/
/* Modified for version 40. This implements local	*/
/* field definitions. Also enumeration values are now	*/
/* all entered by menus.				*/
/*							*/
/* Modified for version 41. Abstraction rules do not	*/
/* relate to fields.					*/
/*							*/
/* Modified for version 42. List matching gives a value	*/
/* which reflects how closely the lists match.		*/
/* similarenum() now only requires 2 parameters, and	*/
/* similarrange() now checks that the fieldtype is	*/
/* NUMTYPE.						*/
/*							*/
/* Modified for version 45. If a field definition has a	*/
/* prompt list associated with it, then this is		*/
/* displayed in lieu of the normal prompt.		*/
/*							*/
/* Modified for version 48. The enterfield(s) routines	*/
/* have been altered to allow prompts to be expanded	*/
/* for ordinary field definitions.			*/
/*							*/
/* Modified for version 49. The pre-processing rules	*/
/* are applied before performing the search.		*/
/*------------------------------------------------------*/

void casematching(casehunt)	/* top level casematching routine */
int casehunt;	/* TRUE if case is used to get index values */
{
 entercase();	/* the user enters a case, usercase points to it */
 preprocess();	/* apply pre-processing rules to usercase */	
 indexsearch(casehunt);	/* perform an index search */
}


/* procedure to allow the user to enter case data to be matched */

void entercase()
{

 usercase = (struct casenode *) calloc(1, sizeof(struct casenode));
				/* create a case node */
 strcpy(usercase->name,"usercase");
 usercase->solnhead = NULL;	/* for now */
 enterfields(&(usercase->fieldhead));	/* enter problem fields only */
 usercase->fielddefhead = NULL;	/* no local field definitions */
 usercase->rephead = NULL;	/* no local repair rules attached */
 usercase->caselink = NULL;	/* this case is not part of a list */
}


/* enterfields allows the user to enter data for the case that is
   to be matched */

void enterfields(headptr)
struct fieldnode **headptr;
{
 enterfield(probdefnroot,headptr);
}


/* procedure to enter a value of a field */

void enterfield(defptr,headptr)
struct typenode *defptr;
struct fieldnode **headptr;
{
struct fieldnode *pointer;

 if (defptr->fieldtype == STRTYPE)
    pointer = enterstrvalue(defptr);	/* enter a string field */
 else if (defptr->fieldtype == NUMTYPE)
    pointer = enternumvalue(defptr);	/* enter a number field */
 else if (defptr->fieldtype == ENUMTYPE)
    pointer = enterenum(defptr);	/* enter enumeration field */
 else
   {
    if (strcmp(defptr->fieldname,"keywords") == 0)
      {
       pointer = entertext(defptr);		/* enter a text list */
      }
    else
      {
       pointer = enterlistvalue(defptr);/* enter a list field */
      }
   }

 if (pointer == NULL)	/* no value was entered for this field */
   {
    if (defptr->typelink == NULL)	/* last field */
      {
       *headptr = NULL;
       return;
      }
    else
      {
/*       *headptr = pointer;
       enterfield(defptr->typelink,&(pointer->fieldlink));*/

       enterfield(defptr->typelink,headptr);
			/* look at next field */
       return;
      }
   }

 if (defptr->typelink == NULL)	/* last field */
   {
    pointer->fieldlink = NULL;
    *headptr = pointer;
   }
 else
   {
    *headptr = pointer;
    enterfield(defptr->typelink,&(pointer->fieldlink));	/* enter next field */
   }
}


/* enter a string value field */

struct fieldnode *enterstrvalue(defptr)
struct typenode *defptr;
{
buftype stringvalue;
char *strptr;

enterstring:

 if (defptr->prompt != NULL)
   {
    fp = stdout;
    expandlist(defptr->prompt,usercase);
    putchar(' ');
   }
 else
   {
    printf("Enter the string value for %s: ",defptr->fieldname);
   }

 getstring(stringvalue);

 if (stringvalue[0] == NULL)	/* null string */
   {
    return NULL;
   }
 else if (strlen(stringvalue) > STRBUFSIZE-1)
   {
    fprintf(stderr,"Error - string is too long\n");
    goto enterstring;
   }
 else
   {
    strptr = createstringnode();   
    strcpy(strptr,stringvalue); /* copy string to string node */
    return createstrfield(defptr->fieldname, strptr);
   }
}


/* enter a number field */ 
 
struct fieldnode *enternumvalue(defptr) 
struct typenode *defptr;   
{ 
buftype stringvalue; 
 
 if (defptr->prompt != NULL)
   {
    fp = stdout;
    expandlist(defptr->prompt,usercase);
    putchar(' ');
   }
 else
   {
    printf("Enter the number value for %s: ",defptr->fieldname); 
   }

 getstring(stringvalue); 

 if (stringvalue[0] == NULL)    /* null string */
    return NULL; 
 else
    return createnumfield(defptr->fieldname, atof(stringvalue));
 
}


/* enter an enumeration value */

struct fieldnode *enterenum(defptr)
struct typenode *defptr;
{
 return enterenummenu(defptr);
}


/* enter an enumeration value using a menu */

struct fieldnode *enterenummenu(defptr)
struct typenode *defptr;
{
int choice;

 displayenummenu(defptr);
 choice = getenumchoice(enumlistlength(defptr->enumhead),
                        defptr->fieldname);
 if (choice == 0)
   {
    return NULL;
   }
 else
   {
    return createenumfield(defptr->fieldname,enumchoice(choice,defptr->enumhead));
   }
}


/* display a menu for entering an enumeration value */

void displayenummenu(defptr)
struct typenode *defptr;
{
int n;
struct enumnode *pointer;

 if (defptr->prompt != NULL)
   {
    fp = stdout;
    expandlist(defptr->prompt,usercase);
    putchar('\n');
   }
 else
   {
    printf("\n\nSelect value for %s\n",defptr->fieldname);
   }

 n = 1;
 pointer = defptr->enumhead;

 while(pointer != NULL)
   {
    printf("\n\t%d) %s",n,pointer->evalue);
    n++;
    pointer = pointer->enumlink;
   }

 printf("\n\n");
}


/* get user's choice for enumeration value */

int getenumchoice(last,fieldname)
int last;
buftype *fieldname;
{
int choice;
char inputstring[40];

 do
   {
    printf("Enter choice (1 to %d",last);
    if (fieldisindexed(fieldname) == TRUE)
      {
       printf("): ");
      }
    else
      {
       printf(" or hit RETURN): ");
      }
    getstring(inputstring);
    if (strcmp(inputstring,"") == 0 &&
        fieldisindexed(fieldname) == FALSE)
      {
       return 0;	/* no input */
      }
    choice = atoi(inputstring);
   }
 while (choice < 1 || choice > last);

 printf("\n\n");
 return choice;
}


/* return a stringnode containing the user's enumeration choice */

char *enumchoice(choice,enumptr)
int choice;
struct enumnode *enumptr;
{
int n;
struct enumnode *pointer;
char *strptr;

 n = 1;
 pointer = enumptr;

 while (n < choice)
   {
    n++;
    pointer = pointer->enumlink;
   }

 strptr = createstringnode();
 strcpy(strptr,pointer->evalue);
 return strptr;
}


/* return TRUE if field is used as an index */

int fieldisindexed(fieldname)
buftype *fieldname;
{
struct indexnamenode *pointer;

 pointer = indexnameroot;

 while (pointer != NULL)
   {
    if (strcmp(pointer->fieldname,fieldname) == 0)
      {
       return TRUE;	/* found matching index name */
      }
    pointer = pointer->indexnamelink;
   }

 return FALSE;
}


/* enter a list */

struct fieldnode *enterlistvalue(defptr)
struct typenode *defptr;
{
 setjmp(environment);

 /* longjmp jumps to here if there was a syntax error */

 if (defptr->prompt != NULL)
   {
    fp = stdout;
    expandlist(defptr->prompt,usercase);
    printf(" use ; to terminate: ");
   }
 else
   {
    printf("Enter the list value for %s, use ; to terminate: ",defptr->fieldname);
   }
 ch = getchar();        /* get 1st character from terminal, to initilise gettok */
 
 gettok();      /* get next token and lexeme values */
 if (token == ';')	/* no list entered */
    return NULL;
 return createlistfield(defptr->fieldname, parsethelist());
}


/* a special procedure to input a string */
void getstring(str)
char *str;
{
char c,*string;

 string = str;

 c = getchar();
 while ( c == ' ' || c == '\t' ) c = getchar();	/* skip white space */

 while (c != '\n')
   {
    *(string++) = c;
    c = getchar();
   }

 *string = '\0';
}

 
/* parse the list input by the user. Return a pointer to the created list */

struct listnode *parsethelist()
{
struct listnode *pointer;

 pointer = parselist();	/* do the parsing */
 match(";");
 return pointer;
}

/* top level parsing routine. retuns a pointer to the created list */

struct listnode *parselist()
{
struct listnode *pointer;

 match("[");	/* match lexeme with "[", error if it doesn't match */
 gettok();	/* get next token */
 pointer = parsesequence();	/* parse a sequence of items */
 match("]");
 gettok();
 return pointer;
}

/* parse the sequence of items inside the square brackets */

struct listnode *parsesequence()
{
struct listnode *pointer;
char *strptr;

 if (strcmp(lexeme,"]") == 0)	/* empty list */
    return NULL;

 if (token == REALNUM)
   {
    pointer = createlistnum(atof(lexeme));
    gettok();
   }
 else if (token == '-')
   {
    gettok();
    if (token == REALNUM)
      {
       pointer = createlistnum(atof(lexeme) * -1.0);
       gettok();
      }
    else
      { 
       fprintf(stderr,"Error - Syntax Error in list\n");
       while (ch != '\n') ch = getchar();
       longjmp(environment,1);
      }
   }
 else if (token == STRING)
   {
    strptr = createstringnode();
    strcpy(strptr,lexeme);
    pointer = createliststring(strptr);
    gettok();
   }
 else if (token == IDENTIFIER)
   {
    strptr = createstringnode();
    strcpy(strptr,lexeme);
    pointer = createlistenum(strptr);
    gettok();
   }
 else if (token == '[')		/* item is a list */
   {
    pointer = createlistlist(parselist());
   }
 else
   {
    fprintf(stderr,"Error - Syntax Error in list\n");
    while (ch != '\n') ch = getchar();
    longjmp(environment,1);
   }


 pointer->listlink = parsesequence();

 return pointer;
}


/* gettok is the lexical analyser for parsing lists */

void gettok()
{
int lindex;	/* index into lexeme */

 lindex = 0;
 while (ch == ' ' || ch == '\t' || ch == '\n') ch = getchar();	/* skip whitespace */

 if (ch == '[' ||
     ch == ']' ||
     ch == ',' ||
     ch == '-' ||
     ch == ';')	/* one character lexemes */
   {
    lexeme[lindex] = ch;
    lexeme[lindex+1] = '\0';
    token = (int) ch;
    ch = getchar();
    return;
   }

 if (ch == '\'')	/* string lexeme */
   {
    ch = getchar();
    while (ch != '\'')
      {
       if (ch == '\n')
         {
          printf("newline character in string\n");
          longjmp(environment,1);
         }
       lexeme[lindex] = ch;
       ch = getchar();
       lindex++;
       if (lindex == STRBUFSIZE-1)
         {
          fprintf(stderr,"Error - string is too long\n");
          while (ch != '\n') ch = getchar();
          longjmp(environment,1);
         }
      }
    lexeme[lindex] = '\0';
    ch = getchar();
    token = STRING;
    return;
   }
 else if (ch >= '0' && ch <= '9') /* lexeme is a real */
   {
    while (ch >= '0' && ch <= '9') /* read in whole part */
      {
       lexeme[lindex++] = ch;
       ch = getchar();
      }
 
    if (ch == '.') /* number has fractional part */
      {
       lexeme[lindex++] = ch;
       ch = getchar();
       while (ch >= '0' && ch <= '9') /* read in fractional part */
         {
          lexeme[lindex++] = ch;
          ch = getchar();
         }
      }
    lexeme[lindex] = '\0'; 
    token = REALNUM;
    return;
   }
 else if ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z')) /* lexeme is identifier*/
   {
    while ((ch >= 'a' && ch <= 'z') ||
           (ch >= 'A' && ch <= 'Z') ||
           (ch >= '0' && ch <= '9') ||
           (ch == '_'))
      {
       if (lindex == STRBUFSIZE-1)
         {
          fprintf(stderr,"Error - symbol is too long\n");
          while (ch != '\n') ch = getchar();
          longjmp(environment,1);
         }
       lexeme[lindex++] = ch;
       ch = getchar();
      }
    lexeme[lindex] = '\0'; /* terminate identifier */
    token = IDENTIFIER;
    return;
   }
 else
   {
    fprintf(stderr,"Error - Syntax Error in list\n");
    while (ch != '\n') ch = getchar();
    longjmp(environment,1);
   }
}


/* match compares the string parameter with lexeme */

void match(str)
char *str;
{
 if (strcmp(lexeme,str) != 0)
   {
    fprintf(stderr,"Error - Syntax Error in list\n");
    while (ch != '\n') ch = getchar();
    longjmp(environment,1);
   }
}


/* return true if enumeration symbol is ok */
int checkenumok(fieldname, evalue, defnroot)
struct typenode *defnroot;
buftype fieldname;
buftype evalue;
{
struct enumnode *head;

 if (abstract(evalue))
   {
    return TRUE;
   }

 head = searchtypelistok(fieldname, defnroot);	/* problem fields only */

 if (head == NULL)
   {
    return FALSE;	/* field does not exist */
   }

 return searchenodesok(head, evalue);
}


/* search for a field definition, return the head of the enumeration list */

struct enumnode *searchtypelistok(fieldname, nodeptr)

buftype fieldname;
struct typenode *nodeptr;
{
struct typenode *pointer;

 pointer = nodeptr;

 while (pointer != NULL)
   {
    if (strcmp(pointer->fieldname, fieldname) == 0) /* found field name */
      {
       return pointer->enumhead; /* return pointer to head of enumeration list*/
      }
    else
      {
       pointer = pointer->typelink;
      }
   }
 printf("Fieldname %s not defined\n",fieldname);
 return NULL;
}
 
 
int searchenodesok(enumptr, evalue) /* search for enumeration string */
struct enumnode *enumptr;
buftype evalue;
{
struct enumnode *pointer;
 
 pointer = enumptr;
 
 while (pointer != NULL)
   {
    if (strcmp(pointer->evalue, evalue) == 0)
      {
       return TRUE; /* symbol exists so finish */
      }
    else
      {
       pointer = pointer->enumlink;
      }
   }
 printf("Enumeration symbol %s not defined\n",evalue);
 return FALSE;
}


/* enter text for keyword matching */

struct fieldnode *entertext(defptr)
struct typenode *defptr;
{
struct fieldnode *pointer;

 if (defptr->prompt != NULL)
   {
    fp = stdout;
    expandlist(defptr->prompt,usercase);
    printf(" (use . to terminate):\n\n");
   }
 else
   {
    printf("\nEnter description (use . to terminate):\n\n");
   }
 ch = getchar();
 pointer = createlistfield("keywords",createtextlist());
 return pointer;
}


/* create a list from entered text */

struct listnode *createtextlist()
{
struct listnode *head,*tail;
buftype buffer, *pointer;
int index;

 while (ch == ' ' ||
        ch == '\n' ||
        ch == '\t' ||
        ch == ',' ||
        ch == ';' ||
	ch == '!' ||
        ch == '?' ||
        ch == '-')
   {
    ch = getchar();	/* skip white space */
   }

 if (ch == '.')
   {
    while ((ch = getchar()) != '\n');
    return NULL;
   }

 index = 0;

 while (index < STRBUFSIZE - 1 &&	/* put characters into buffer */
        ch != ' ' &&
        ch != '\n' &&
	ch != '.' &&
        ch != ',' &&
        ch != ';' &&
        ch != '!' &&
        ch != '?' &&
        ch != '-' &&
        ch != '\t')
   {
    buffer[index++] = isupper(ch) ? tolower(ch) : ch;	/* convert to lower case */
    ch = getchar();
   }

 if (index == STRBUFSIZE - 1 &&
     ch != ' ' && 
     ch != '\n' &&
     ch != '.' &&
     ch != ',' &&
     ch != ';' &&
     ch != '!' &&
     ch != '?' &&
     ch != '-' &&
     ch != '\t')
   {
    printf("\nWarning - word is too long\n\n");
    while (ch != ' ' &&
           ch != '\n' &&
           ch != '\t' &&
           ch != '.' &&
           ch != ',' &&
           ch != '!' &&
           ch != '?' &&
           ch != '-' &&
           ch != ';')
      {
       ch = getchar();	/* ignore remainder of word */
      }
   }

 buffer[index] = '\0';	/* terminate string */

 if (strcmp(buffer,"a") == 0 ||		/* strip out non-words */
     strcmp(buffer,"ah") == 0 ||
     strcmp(buffer,"ahem") == 0 ||
     strcmp(buffer,"am") == 0 ||
     strcmp(buffer,"an") == 0 ||
     strcmp(buffer,"any") == 0 ||
     strcmp(buffer,"are") == 0 ||
     strcmp(buffer,"as") == 0 ||
     strcmp(buffer,"at") == 0 ||
     strcmp(buffer,"bah") == 0 ||
     strcmp(buffer,"be") == 0 ||
     strcmp(buffer,"but") == 0 ||
     strcmp(buffer,"by") == 0 ||
     strcmp(buffer,"eg") == 0 ||
     strcmp(buffer,"eh") == 0 ||
     strcmp(buffer,"etc") == 0 ||
     strcmp(buffer,"for") == 0 ||
     strcmp(buffer,"gee") == 0 ||
     strcmp(buffer,"ha") == 0 ||
     strcmp(buffer,"heh") == 0 ||
     strcmp(buffer,"hey") == 0 ||
     strcmp(buffer,"ho") == 0 ||
     strcmp(buffer,"ie") == 0 ||
     strcmp(buffer,"is") == 0 ||
     strcmp(buffer,"it") == 0 ||
     strcmp(buffer,"its") == 0 ||
     strcmp(buffer,"it's") == 0 ||
     strcmp(buffer,"lo") == 0 ||
     strcmp(buffer,"of") == 0 ||
     strcmp(buffer,"oh") == 0 ||
     strcmp(buffer,"ooh") == 0 ||
     strcmp(buffer,"so") == 0 ||
     strcmp(buffer,"that") == 0 ||
     strcmp(buffer,"the") == 0 ||
     strcmp(buffer,"there") == 0 ||
     strcmp(buffer,"these") == 0 ||
     strcmp(buffer,"this") == 0 ||
     strcmp(buffer,"those") == 0 ||
     strcmp(buffer,"to") == 0 ||
     strcmp(buffer,"um") == 0 ||
     strcmp(buffer,"umm") == 0 ||
     strcmp(buffer,"which") == 0 ||
     strcmp(buffer,"with") == 0)
   {
    return createtextlist();
   }
 else
   {
    pointer = createbuftypenode();
    strcpy(pointer,buffer);
    tail = createtextlist();
    head = createliststring(pointer);	/* get next word */
    head->listlink = tail;
    return head;
   }
}


/*------------------------------------------------------*/
/*							*/
/*		The really interesting bit		*/
/*							*/
/*------------------------------------------------------*/

/* main search procedure */

void indexsearch(casehunt)
int casehunt;		/* TRUE if case is used for index values */
{
struct indexspecnode *indexspec;
int getindexes;

 getindexes = !casehunt;	/* NOT operation */

 do
   {
    if (getindexes == TRUE)
      {
       indexspec = specifyindexes();	/* get index values from user */
      }
    else
      {
       indexspec = makeindexspec(indexnameroot);
      }
    performsearch(indexspec);	/* perform the search */

    if (caseptrroot == NULL)
      {
       printf("\n\nNo cases matched all index constraints...\n\n");
       getindexes = TRUE;
      }
    deleteindexspec(indexspec);
   }
 while (caseptrroot == NULL);

 if (autoselect() == TRUE)
   {
    foundcase = matchcase(usercase);
   }
 else
   {
    do
      {
       foundcase = browsecaselist(caseptrroot);
      }
    while (foundcase == NULL);
   }
}


/* return TRUE if user wants to auto select case */

int autoselect()
{
char reply[10];

 printf("\n\n\n");
 do
   {
    printf("Use weights to select case (yes/no)? ");
    getstring(reply);
   }
 while (strcmp(reply,"yes") != 0 && strcmp(reply,"no") != 0);

 if (strcmp(reply,"yes") == 0)
   {
    return TRUE;
   }
 else
   {
    return FALSE;
   }
}


/* get index specification from usercase */

struct indexspecnode *makeindexspec(indexnameptr)
struct indexnamenode *indexnameptr;
{
struct indexspecnode *tail,*pointer;
struct fieldnode *fieldptr;

 if (indexnameptr == NULL)
   {
    return NULL;
   }
 else
   {
    tail = makeindexspec(indexnameptr->indexnamelink);
    fieldptr = searchfield(indexnameptr->fieldname,usercase->fieldhead);
    pointer = createindexspecnode(fieldptr->fieldname,fieldptr->strcontents);
    pointer->indexspeclink = tail;
    return pointer;
   }
}


/* delete index specification list */

void deleteindexspec(indexspec)
struct indexspecnode *indexspec;
{
 if (indexspec == NULL)
   {
    return;
   }
 else
   {
    deleteindexspec(indexspec->indexspeclink);
    free(indexspec);
   }
}


/* get the indexes & values to search for from user */

struct indexspecnode *specifyindexes()
{
struct indexspecnode *indexspec,*pointer;
buftype fieldname,value;
int enumok;

 indexspec = NULL;

 do
   {
    printf("\n\n\nSpecify index constraints:\n\n");
    do
      {
       do
         {
          printf("Enter an index fieldname (or hit RETURN): ");
          getstring(fieldname);
          if (strcmp(fieldname,"") == 0)
             break;
         }
       while(fieldisindexed(fieldname) == FALSE);

       if (strcmp(fieldname,"") == 0)
       break;

       do
         {
          printf("Enter the value for the index field: ");
          getstring(value);
          enumok = checkenumok(fieldname,value,probdefnroot);
         }
       while (enumok == FALSE);

       pointer = createindexspecnode(fieldname,value);
       pointer->indexspeclink = indexspec;
       indexspec = pointer;
      }
    while (strcmp(fieldname,"") != 0);
   }
 while (indexspec == NULL);

 return indexspec;
}


/* perform the index search */

void performsearch(indexspec)
struct indexspecnode *indexspec;
{
struct indexspecnode *pointer;

 performfirstpass(indexspec);	/* search on first index */

 pointer = indexspec->indexspeclink;	/* perform search on tail of indexspec */

 while (pointer != NULL)
   {
    performnextpass(pointer);
    pointer = pointer->indexspeclink;
   }
}


/* get the initial set of cases matching the first index specification */

void performfirstpass(indexspec)
struct indexspecnode *indexspec;
{
 caseptrroot = buildmatchlist(indexspec);	/* create list of matching cases */
}


/* get the next set of matching cases and perform an intersection */

void performnextpass(indexspec)
struct indexspecnode *indexspec;
{
struct caseptrnode *matches;

 matches = buildmatchlist(indexspec);	/* create list of cases matching this spec */
 intersection(matches);	/* intersection with cases matched so far */
 deletecaseptrlist(matches);
}


/* build a list of case pointers to matching cases */

struct caseptrnode *buildmatchlist(indexspec)
struct indexspecnode *indexspec;
{
struct indexnamenode *indexname;
struct indexnode *index,*foundindex;
struct caseptrnode *list,*specialisations;

 indexname = findindexname(indexspec->fieldname);
 index = indexname->indexhead;
 foundindex = findindex(index,indexspec->value);

 if (foundindex == NULL)
   {
    list = NULL;
   }
 else
   {
    list = copycaseptrlist(foundindex->caseptrhead);
   }

 if (abstract(indexspec->value) == TRUE)
   {						/* if symbol is abstract */
    specialisations = searchabstract(indexspec);
    getunion(&list,specialisations);	/* return the specialisations as well */
    deletecaseptrlist(specialisations);
    return list;
   }
 else
   {
    return list;
   }
}


/* search index node list for matching value */

struct indexnode *findindex(indexhead,value)
struct indexnode *indexhead;
buftype *value;
{
struct indexnode *pointer;

 pointer = indexhead;

 while (pointer != NULL)
   {
    if (strcmp(pointer->evalue,value) == 0)
      {
       return pointer;
      }
    pointer = pointer->indexlink;
   }

 return NULL;
}


/* perform search on enumerations corresponding to abstract index */

struct caseptrnode *searchabstract(indexspec)
struct indexspecnode *indexspec;
{
struct caseptrnode *pointer,*temp;
struct modnode *modptr;

 pointer = NULL;
 modptr = modroot;

 while (modptr != NULL)
   {
    if (modptr->fieldtype == ENUMTYPE)
      {
       if (strcmp(modptr->abstraction,indexspec->value) == 0)
         {
          temp = matchenumlist(indexspec->fieldname,modptr);
          getunion(&pointer,temp);/* add cases matching enumeration list */
          deletecaseptrlist(temp);
         }
      }
    modptr = modptr->modlink;
   }

 return pointer;
}


/* make a copy of a case pointer list */

struct caseptrnode *copycaseptrlist(caseptrlist)
struct caseptrnode *caseptrlist;
{
struct caseptrnode *pointer,*tail;

 if (caseptrlist == NULL)	/* null list */
   {
    return NULL;
   }
 else
   {
    tail = copycaseptrlist(caseptrlist->caseptrlink);
    pointer = createcaseptrnode(caseptrlist->caseind);
    pointer->caseptrlink = tail;
    return pointer;
   }
}


/* perform an intersection between the cases matched
   so far and a case pointer list */

void intersection(caseptrlist)
struct caseptrnode *caseptrlist;
{
struct caseptrnode *pointer,*temp;

 pointer = caseptrroot;

 while (pointer != NULL)
   {
    if (caseptrinlist(pointer->caseind,caseptrlist) == FALSE)
      {
       temp = pointer->caseptrlink;
       deletecaseptrnode(pointer);	/* if index not in both lists, delete */
       pointer = temp;
       continue;
      }
    pointer = pointer->caseptrlink;
   }
}


/* perform a union of two lists */

void getunion(firstlistptr, secondlist)
struct caseptrnode **firstlistptr;
struct caseptrnode *secondlist;
{
struct caseptrnode *pointer;

 pointer = secondlist;

 while (pointer != NULL)
   {
    if (caseptrinlist(pointer->caseind,*firstlistptr) == FALSE)
      {
       addcaseptrtolist(pointer->caseind,firstlistptr);	/* eliminates duplicates */
      }
    pointer = pointer->caseptrlink;
   }

}


/* build list of matching cases for symbols in enumeration list */

struct caseptrnode *matchenumlist(fieldname,modptr)
struct modnode *modptr;
buftype *fieldname;
{
struct enumnode *enumptr;
struct indexspecnode *tempindexspec;
struct caseptrnode *pointer,*temp;

 tempindexspec = createindexspecnode(fieldname,"");
 pointer = NULL;
 enumptr = modptr->enumhead;

 while (enumptr != NULL)
   {
    strcpy(tempindexspec->value,enumptr->evalue);
    temp = buildmatchlist(tempindexspec);
    getunion(&pointer,temp);
    deletecaseptrlist(temp);
    enumptr = enumptr->enumlink;
   }

 free(tempindexspec);
 return pointer;
}


/* return TRUE if a case pointer is in a case pointer list */

int caseptrinlist(caseind,caseptrlist)
struct casenode *caseind;
struct caseptrnode *caseptrlist;
{
struct caseptrnode *pointer;

 pointer = caseptrlist;

 while (pointer != NULL)
   {
    if (caseind == pointer->caseind)
      {
       return TRUE;
      }
    pointer = pointer->caseptrlink;
   }

 return FALSE;
}


/* delete a case pointer node from the main list */

void deletecaseptrnode(caseptr)
struct caseptrnode *caseptr;
{
struct caseptrnode *pointer;

 if (caseptrroot == NULL)
   {
    fprintf(stderr,"Huh? - deleting from null case pointer list\n");
    exit(1);	/* this won't happen */
   }

 if (caseptr == caseptrroot)	/* deleting first item in list */
   {
    caseptrroot = caseptrroot->caseptrlink;
    free(caseptr);
   }
 else	/* delete from tail of list */
   {
    pointer = caseptrroot;

    while (pointer->caseptrlink != NULL)
      {
       if (pointer->caseptrlink == caseptr)
         {
          pointer->caseptrlink = caseptr->caseptrlink;
          free(caseptr);
          return;
         }
       pointer = pointer->caseptrlink;
      }

    fprintf(stderr,"Huh? - deleting non-existant pointer from case pointer list\n");
    exit(1);
   }
}


/* add a case pointer to a list */

void addcaseptrtolist(caseind,caseptrlist)
struct casenode *caseind;
struct caseptrnode **caseptrlist;
{
struct caseptrnode *pointer;

 pointer = createcaseptrnode(caseind);
 pointer->caseptrlink = *caseptrlist;
 *caseptrlist = pointer;
}


/* return pointer to indexnamenode for named field */

struct indexnamenode *findindexname(fieldname)
buftype *fieldname;
{
struct indexnamenode *indexnameptr;

 indexnameptr = indexnameroot;

 while (indexnameptr != NULL)
   {
    if (strcmp(indexnameptr->fieldname,fieldname) == 0)
      {
       return indexnameptr;
      }
    indexnameptr =indexnameptr->indexnamelink;
   }

 fprintf(stderr,"Huh? - field %s should be indexed\n",fieldname);
 exit(1);
}


/* browse through list of selected cases */
/* Select one for repair */

struct casenode *browsecaselist(caseptrlist)
struct caseptrnode *caseptrlist;
{
char choice[10];
struct casenode *tailchoice;

displayagain:

 printf("\n\n\n");
 fp = stdout;
 displaycase(caseptrlist->caseind);
 printf("\n\n");

 do
   {
    printf("S = Select the displayed case\nB = Go Back to previous case\n");
    if (caseptrlist->caseptrlink != NULL)
      {
       printf("N = look at Next case\n");
      }
    printf("\nEnter choice: ");
    getstring(choice);
    choice[0] = toupper(choice[0]);
   }
 while ((caseptrlist->caseptrlink == NULL &&
         choice[0] != 'S' &&
         choice[0] != 'B') ||
        (caseptrlist->caseptrlink != NULL &&
         choice[0] != 'S' &&
         choice[0] != 'B' &&
         choice[0] != 'N'));

 if (choice[0] == 'S')
    return caseptrlist->caseind;
 else if (choice[0] == 'B')
    return NULL;
 else
   {
    tailchoice = browsecaselist(caseptrlist->caseptrlink);
    if (tailchoice == NULL)
      {
       goto displayagain;	/* oh dear, oh dear */
      }
    else
      {
       return tailchoice;
      }
   }
}


struct casenode *matchcase(compcase)
struct casenode *compcase;
{
 return findcase(caseptrroot,compcase);
}

/* findcase compares the case with the current one in the list and
   finds out the number of matching fields. It then calls itself
   recursively, if no case in the tail of the list has a greater
   number of matching fields then the current case is returned */

struct casenode *findcase(listcaseptr,compcase)
struct casenode *compcase;
struct caseptrnode *listcaseptr;
{
float totalweight,bestweight;
struct casenode *bestcase;

 bestweight = -1000000.0;
 bestcase = listcaseptr->caseind;

 while (listcaseptr != NULL)
   {		/* get the total weight for the current case */
    totalweight = findmatchfields(listcaseptr->caseind,compcase);

    if (totalweight > bestweight)	/* this one is a better match */
      {
       bestweight = totalweight;
       bestcase = listcaseptr->caseind;
      }

    listcaseptr = listcaseptr->caseptrlink;
   }

 return bestcase;
}


/* find the total weight of matching fields for the current case */
float findmatchfields(listcase,compcase)
struct casenode *listcase,*compcase;
{
float totalweight;
struct fieldnode *listptr,*compptr;

 totalweight = 0.0;
 compptr = compcase->fieldhead;	/* points to first field in compare case */

 while (compptr != NULL)
   {
    listptr = listcase->fieldhead;	/* first field in list case */
    while (listptr != NULL)
      {
       if (strcmp(compptr->fieldname,listptr->fieldname) == 0)	/* fieldnames match */
         {
          totalweight += domatch(listptr,compptr);	/* domatch returns weight if the fields match*/
          break;
         }
       listptr = listptr->fieldlink;
      }
    compptr = compptr->fieldlink;
   }

 /* now compare any anonymous fields with any fields in the solution
    part whose names match. */

 compptr = compcase->fieldhead;

 while (compptr != NULL)
   {
    listptr = listcase->solnhead;	/* first solution field in list case */
    while (listptr != NULL)
      {
       if (anonymous(compptr->fieldname,compptr) &&
           strcmp(compptr->fieldname,listptr->fieldname) == 0)
			/* anonymous field matches solution field */
         {
printf("anonymous field matches solution fieldname: %s\n",
            compptr->fieldname);

          totalweight += domatch(listptr,compptr);
          break;
         }
       listptr = listptr->fieldlink;
      }
    compptr = compptr->fieldlink;
   }

 return totalweight;
}


/* function returns TRUE if the named field is anonymous */

int anonymous(fieldname,caseptr)
buftype *fieldname;
struct casenode *caseptr;
{
 if (
     (findfield(fieldname,probdefnroot) == NULL) &&
           ((caseptr->fielddefhead == NULL) ||
            (caseptr->fielddefhead != NULL &&
            findfield(fieldname,caseptr->fielddefhead) == NULL))
    )
   {
    return TRUE;
   }
 else
   {
    return FALSE;
   }
}


/* compare two fields, return weight if they match, 0 otherwise */

float domatch(listptr,compptr)
struct fieldnode *listptr,*compptr;
{
 if (listptr->fieldtype == STRTYPE)
    return compstrfields(listptr,compptr);	/* compare string fields */
 else if (listptr->fieldtype == ENUMTYPE)
    return compenumfields(listptr,compptr);	/* compare enumeration fields */
 else if (listptr->fieldtype == NUMTYPE)
    return compnumfields(listptr,compptr);	/* compare number fields */
 else if (listptr->fieldtype == LISTTYPE)
    return complistfields(listptr,compptr);	/* compare list fields */
 else
   {
    fprintf(stderr,"Huh? - Unknown type comparing field %s\n",listptr->fieldname);
    exit(1);
   }
}


/* compare two string fields. return weight if they match, 0 otherwise */

float compstrfields(listptr,compptr)
struct fieldnode *listptr,*compptr;
{
 if (strcmp(listptr->strcontents,compptr->strcontents) == 0)
    return findweight(listptr->fieldname);
 else
    return 0.0;
}


/* compare two enumeration values. return weight if they match, 0 otherwise */
 
float compenumfields(listptr,compptr)
struct fieldnode *listptr,*compptr;
{
 if (strcmp(listptr->strcontents,compptr->strcontents) == 0)
   {
    return findweight(listptr->fieldname);
   }
 else if (similarenum(listptr->strcontents,compptr->strcontents))
   {
    return SIMWEIGHT * findweight(listptr->fieldname);
   }
 else
   {
    return 0.0;
   }
}


/* compare two number fields. return weight if they match, 0 otherwise */

float compnumfields(listptr,compptr)
struct fieldnode *listptr,*compptr;
{
 if (listptr->numcontents == compptr->numcontents)
   {
    return findweight(listptr->fieldname); 
   }
 else if (similarrange(listptr->numcontents,compptr->numcontents,listptr->fieldname))
   {
    return ( 1 -
             (
               (fabs(compptr->numcontents - listptr->numcontents)) / 
               (fabs(upperbound - lowerbound))
             )
           )
             * findweight(listptr->fieldname);
   }
 else 
   {
    return 0.0; 
   }
}


/* compare two list fields. return weight if they match, 0 otherwise */

float complistfields(listptr,compptr)
struct fieldnode *listptr,*compptr;
{
 if (strcmp(listptr->fieldname,"keywords") == 0)
    return matchkeys(listptr->listcontents, compptr->listcontents);
 else
    return matchlist(findweight(listptr->fieldname),
                     listptr->listcontents,
                     compptr->listcontents);
}


/* return weight of matching keywords in list */

float matchkeys(listptr,compptr)
struct listnode *listptr, *compptr;
{
struct listnode *pointer;
char buffer[STRBUFSIZE * 3];
float matches,length;

 length = listlength(listptr);

 if (length == 0.0)
   {
    return 0.0;
   }


 matches = 0.0;
 pointer = compptr;

 while (pointer != NULL && matches < listlength(listptr))
			/* for each string in user's list */
   {
    if (inlist(pointer->strcontents,listptr))	/* string is in keywords list */
      {
       matches++;
      }
    if (pointer->listlink != NULL)	/* join 2 words and try match */
      {
       strcpy(buffer,pointer->strcontents);
       strcat(buffer,(pointer->listlink)->strcontents);
       if (inlist(buffer,listptr))
         {
          matches++;
         }
       if ((pointer->listlink)->listlink != NULL)
         {
          strcat(buffer,((pointer->listlink)->listlink)->strcontents);
          if (inlist(buffer,listptr))
            {
             matches++;
            }
         }
      }
    pointer = pointer->listlink;
   }

 return (matches/length) * findweight("keywords");
}


/* returns the number of items in a list */

int listlength(listptr)
struct listnode *listptr;
{
 if (listptr == NULL)
   {
    return 0;
   }
 else
   {
    return 1 + listlength(listptr->listlink);
   }
}


/* returns TRUE if the string is in the list of strings*/

int inlist(string, listptr)
buftype *string;
struct listnode *listptr;
{
 if (listptr == NULL)	/* reached end of list */
   {
    return FALSE;
   }
 else if (matchwords(string, listptr->strcontents) == TRUE)
   {
    return TRUE;	/* string matches */
   }
 else
   {
    return inlist(string, listptr->listlink);	/* search tail */
   }
}


/* spelling-error tolerant word matching, returns TRUE or FALSE */

int matchwords(str1,str2)
char *str1,*str2;
{
char buf1[3 * STRBUFSIZE],buf2[3 * STRBUFSIZE];

 copyword(buf1,str1);
 copyword(buf2,str2);	/* copy strings to buffers, ignoring common letters */

 juxtaposed(buf1,buf2);	/* correct juxtaposed letters */

 if (strcmp(buf1,buf2) == 0)
   {
    return TRUE;
   }
 else
   {
    return FALSE;
   }
}


/* copy str2 to str1, ignoring common letters */

void copyword(str1, str2)
char *str1,*str2;
{
char *ptr1,*ptr2;
int nastieflag;

 nastieflag = FALSE;	/* set to true when previous letter was common */
 ptr1 = str1;
 ptr2 = str2;

 while (*ptr2 != '\0')
   {
    if (inword(*ptr2,"nastie") == TRUE)	/* common letter */
      {
       ptr2++;	/* so don't copy it */
       nastieflag = TRUE;
       continue;
      }
    else if (ptr1 > str1 && *(ptr1 - 1) == *ptr2 &&	/* duplicate of previous letter */
             nastieflag == FALSE)	/* and no eliminated letters in between */
      {
       ptr2++;	/* so don't copy it */
       continue;
      }
    else
      {
       *(ptr1++) = *(ptr2++);	/* copy and move along to next character */
       nastieflag = FALSE;
      }
   }

 *ptr1 = '\0';	/* terminate word */
}


/* correct any juxtaposed letters in str1 */

void juxtaposed(str1,str2)
char *str1,*str2;
{
char *ptr1,*ptr2,*ptr3,*ptr4,temp;

 if (strlen(str1) < 2 || strlen(str2) < 2)
   {
    return;	/* only one letter in string */
   }

 ptr1 = str1;
 ptr2 = str1 + 1;
 ptr3 = str2;
 ptr4 = str2 + 1;

 while (*ptr2 != '\0' && *ptr4 != '\0')
   {
    if (*ptr1 == *ptr4 && *ptr2 == *ptr3)	/* found juxtaposed letters */
      {
       temp = *ptr1;
       *ptr1 = *ptr2;
       *ptr2 = temp;		/* swap letters over in str1 */
      }
    ptr1++;	/* move to next letter up */
    ptr2++;
    ptr3++;
    ptr4++;
   }
}


/* return TRUE if the letter is in the string */

int inword(chr,string)
char chr,*string;
{
char *ptr;

 ptr = string;

 while (*ptr != '\0')
   {
    if (*ptr == chr)
      {
       return TRUE;
      }
    else
      {
       ptr++;
      }
   }

 return FALSE;
}


/* return TRUE if the two enumeration symbols are similar */

int similarenum(sym1,sym2)
char *sym1,*sym2;
{
struct modnode *modptr;
struct enumnode *enumptr;

 if (abstract(sym1) ||
     abstract(sym2))
   {
    return FALSE;	/* return false if either symbol is an abstraction */
   }

 modptr = modroot;

 while (modptr != NULL)
   {
    if (modptr->fieldtype == ENUMTYPE)
      {
       enumptr = modptr->enumhead;	/* search list for sym1 */

       while (enumptr != NULL)
         {
          if (strcmp(enumptr->evalue,sym1) == 0)	/* found sym1 */
             break;
          enumptr = enumptr->enumlink;
         }
       if (enumptr == NULL)	/* sym1 was not found */
         {
          modptr = modptr->modlink;	/* look at next rule */
          continue;
         }

       enumptr = modptr->enumhead;	/* search list for sym2 */

       while (enumptr != NULL)
         { 
          if (strcmp(enumptr->evalue,sym2) == 0)	/* found sym2 */ 
             break; 
          enumptr = enumptr->enumlink;
         } 
       if (enumptr != NULL)	/* sym2 was also found */ 
         {
          return TRUE;
         }
      }
    modptr = modptr->modlink;	/* look at next rule */
   }

 return FALSE;
}


/* return TRUE if the symbol is an abstraction symbol */

int abstract(evalue)
buftype *evalue;
{
struct modnode *modptr;

 modptr = modroot;

 while (modptr != NULL)
   {
    if (strcmp(modptr->abstraction, evalue) == 0)/* matches abstraction value */
      {
       return TRUE;
      }
    modptr = modptr->modlink;	/* look at next rule */
   }

 return FALSE;
}


/* return TRUE if the two values are similar */

int similarrange(num1,num2,fieldname)
float num1,num2;
char *fieldname;
{
struct modnode *modptr;

 modptr = modroot;

 while (modptr != NULL)
   {
    if (modptr->fieldtype == NUMTYPE &&
        strcmp(modptr->fieldname, fieldname) == 0)	/* found a matching fieldname */
      {
       if (num1 >= modptr->start && num1 <= modptr->end &&
           num2 >= modptr->start && num2 <= modptr->end)
         {
          upperbound = modptr->end;
          lowerbound = modptr->start;
          return TRUE;	/* both numbers in similar range */
         }
      }
    modptr = modptr->modlink;
   }

 return FALSE;
}


/* find the weight of the named field */

float findweight(name)
char *name;
{
struct typenode *pointer;

 pointer = probdefnroot;

 while (pointer != NULL)
   {
    if (strcmp(pointer->fieldname,name) == 0)
       return pointer->weight;
    pointer = pointer->typelink;
   }

 /* anonymous field, so return default weight */

 return 1.0;
}


/* comparse two lists, return a number which reflects
   what proportion of items of the first list are in the second.
   Sublists which are similar are only matched against once */

float matchlist(weight,list1,list2)
struct listnode *list1,*list2;
float weight;
{
float matched,sublist,length;
struct listnode *pointer1,*pointer2;

 length = listlength(list2);

 if (length == 0.0)
   {
    return 0.0;
   }


 matched = 0.0;
 pointer1 = list1;

 while (pointer1 != NULL)
   {
    pointer2 = list2;
    while (pointer2 != NULL)
      {
       if (pointer1->listtype == pointer2->listtype)
         {
          if (pointer1->listtype == STRTYPE)
            {
             if (strcmp(pointer1->strcontents,pointer2->strcontents) == 0)
               {
                matched += 1.0;
                break;
               }
            }
          else if (pointer1->listtype == NUMTYPE)
            {
             if (pointer1->numcontents == pointer2->numcontents)
               {
                matched += 1.0;
                break;
               }
            }
          else if (pointer1->listtype == ENUMTYPE)
            {
             if (strcmp(pointer1->strcontents,pointer2->strcontents) == 0)
               {
                matched += 1.0;
                break;
               }
             else if (similarenum(pointer1->strcontents,pointer2->strcontents))
               {
                matched += SIMWEIGHT;
                break;
               }
             else if (matchfieldname(pointer1->strcontents,pointer2->strcontents))
               {
                matched += SIMWEIGHT;
                break;
               }
            }
          else if (pointer1->listtype == LISTTYPE)
            {
             sublist = matchsublist(1.0,pointer1->listitem,pointer2->listitem);
             if (sublist > 0.0)
               {
                matched += sublist;
                break;
               }
            }
          else
            {
             fprintf(stderr,"Huh? - Unknown type comparing lists.\n");
             exit(1);
            }
         }
       pointer2 = pointer2->listlink;
      }
    pointer1 = pointer1->listlink;
   }


 /* work out a weight value which reflects how closely the lists match */

 return (matched / length) * weight;
}


float matchsublist(weight,list1,list2)
struct listnode *list1,*list2;
float weight;
{
float matched,sublist,length;
struct listnode *pointer1,*pointer2;

 length = listlength(list1);

 if (length == 0.0)
   {
    return 0.0;
   }


 matched = 0.0;
 pointer1 = list1;
 pointer2 = list2;

 while (pointer1 != NULL && pointer2 != NULL)
   {
    if (pointer1->listtype == pointer2->listtype)
      {
       if (pointer1->listtype == STRTYPE)
         {
          if (strcmp(pointer1->strcontents,pointer2->strcontents) == 0)
             matched++;
          else
            {
             break;
            }
         }
       else if (pointer1->listtype == NUMTYPE)
         {
          if (pointer1->numcontents == pointer2->numcontents)
             matched++;
          else
            {
             break;
            }
         }
       else if (pointer1->listtype == ENUMTYPE)
         {
          if (strcmp(pointer1->strcontents,pointer2->strcontents) == 0)
             matched++;
          else if (similarenum(pointer1->strcontents,pointer2->strcontents))
             matched += SIMWEIGHT;
          else if (matchfieldname(pointer1->strcontents,pointer2->strcontents))
             matched += SIMWEIGHT;
          else
            {
             break;
            }
         }
       else if (pointer1->listtype == LISTTYPE)
         {
          sublist = matchsublist(1.0,pointer1->listitem,pointer2->listitem);
          if (sublist > 0.0)
            {
             matched += sublist;
            }
          else
            {
             break;
            }
         }
       else
         {
          fprintf(stderr,"Huh? - Unknown type comparing lists.\n");
          exit(1);
         }
      }
    else
      {
       break;
      }
    pointer2 = pointer2->listlink;
    pointer1 = pointer1->listlink;
   }

 /* work out a weight value which reflects how closely the lists match */

 return (matched / length) * weight;
}


int matchfieldname(fieldname,contents)
char *fieldname,*contents;
{
struct fieldnode *fieldptr;

	/* find field in usercase */

 fieldptr = searchfield(fieldname,usercase->fieldhead);

 if (fieldptr == NULL || fieldptr->fieldtype != STRTYPE)
   {
    return FALSE;
   }
 else
   {
    if (strcmp(contents,fieldptr->strcontents) == 0)
      {
       return TRUE;
      }
    else
      {
       return FALSE;
      }
   }
}
