
# line 455 "cases49"
#include <stdio.h>
#include <setjmp.h>
#include <alloc.h>
#include <setjmp.h>
#include <math.h>
#include <string.h>
#include "cbrdefs.h"
#define malloc farmalloc
#define calloc farcalloc
#define free farfree

# line 464 "cases49"
typedef union 
  {
   int comptype;	/* comparison operator type */
   float realval; /* real value */
   char *strval; /* string value or enumeration symbol */
   struct fieldnode *fieldptr; /* pointer to field node */
   struct casenode *caseptr; /* pointer to case node */
   struct enumnode *enumptr; /* pointer to enumeration node */
   struct typenode *typeptr; /* pointer to type node */
   struct listnode *listptr; /* pointer to list node */
   struct modnode *modptr; /* pointer to modification node */
   struct repnode *repptr; /* pointer to repair rule node */
   struct replistnode *replistptr; /* pointer to repair action list */
   struct arithnode *arithptr;	/* pointer to arithmetic tree node */
   struct combnode *combptr;	/* pointer to combination node */
   struct combtreenode *combtreeptr;	/* pointer to combination tree */
   struct indexnamenode *indexnameptr;	/* pointer to indexname node */
   struct intronode *introptr;	/* pointer to introduction text node */
   struct listopnode *listopptr;	/* pointer to list operation node */
  } YYSTYPE;
# define STRING 257
# define REALNUM 258
# define NOTEQUAL 259
# define LESSOREQUAL 260
# define MOREOREQUAL 261
# define ABSTRACTION 262
# define CASE 263
# define CHANGE 264
# define DEFINITION 265
# define END 266
# define EVALUATE 267
# define EXTRACT 268
# define FIELD 269
# define FROM 270
# define HD 271
# define INDEX 272
# define INSTANCE 273
# define INTRODUCTION 274
# define IS 275
# define LIST 276
# define LOCAL 277
# define MODIFICATION 278
# define NUMBER 279
# define OF 280
# define ON 281
# define PR 282
# define PREPROCESS 283
# define PROMPT 284
# define RANGE 285
# define REPAIR 286
# define RESELECT 287
# define RULE 288
# define KSTRING 289
# define SIMILAR 290
# define SOLUTION 291
# define THEN 292
# define TL 293
# define TO 294
# define TYPE 295
# define UNDEFINED 296
# define WEIGHT 297
# define WHEN 298
# define IDENTIFIER 299
# define UNARYMINUS 300
# define OR 301
# define AND 302
# define NOT 303
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
extern int yychar;
extern int yyerrflag;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif
YYSTYPE yylval, yyval;
# define YYERRCODE 256

# line 1420 "cases49"


#include "lex.c"

main(argc,argv)
int argc;
char *argv[];
{

 initialise();	/* initialise global variables */

 if (argc != 2)
   {
    printf("Usage : caspian filename\n");
    exit(1);
   }

 strcpy(casefile,argv[1]);	/* copy filename to global variable */

 fp = fopen(argv[1],"r");	/* open the case file */

 if (fp == NULL)	/* file not found */
   {
    printf("Filename: %s not found\n\n",argv[1]);
    exit(1);
   }

/* display a general message */

 displayblurb();

/* parse the case file, perform type checking, display the contents */

 printf("Reading case file...");
 yyparse();

 fclose(fp);	/* close the case file, tidy like */

 printf("ok\n");
 printf("Type checking cases...");
 typecheckfields();
 printf("ok\n");
 printf("Checking for doubly defined fields...");
 doublecheckfields(probdefnroot);
 printf("ok\n");
 printf("Checking local field definitions...");
 checklocalfielddefs();
 printf("ok\n");
 printf("Checking for multiple uses of fields in cases...");
 multiplecheckfields();
 printf("ok\n");
 printf("Checking for doubly defined case names...");
 doublecheckcases();
 printf("ok\n");
 printf("Checking for non-unique enumeration symbols...");
 doublecheckenums(probdefnroot);
 printf("ok\n");
 printf("Checking index block...");
 checkindexblock();
 printf("ok\n");
 printf("Checking modification rules...");
 checkmodrules();
 printf("ok\n");
 printf("Checking pre-processing rules...");
 checkreprules(preprocessroot,NULL);
 printf("\n   checking for reselect...");
 checkreselect();
 printf("ok\n");
 printf("Checking global repair rules...");
 checkreprules(reproot,NULL);
 printf("ok\n");
 printf("Checking local repair rules...");
 checklocalreprules();
 printf("ok\n");
 printf("Building index structure...");
 buildindexes();
 printf("ok\n");
 fp = stdout;	/* output to screen */
 displayintro();	/* display introduction text */
 menu();	/* main menu */
}


/* initialise global variables */

void initialise()
{
 listexpansion = FALSE;
 START = TRUE;
 introroot = NULL;
 probdefnroot = NULL;
 caseroot = NULL;
 caseptrroot = NULL;
 usercase = NULL;
 foundcase = NULL;
 repcase = NULL;
 modroot = NULL;
 reproot = NULL;
 indexnameroot = NULL;
 lines = 1;
}


/* display the inital message */

void displayblurb()
{
 printf("\n\nCASPIAN (CASe Parser by IAN) ver 1.1");
 printf("\n\n(C) Copyright 1996 U.W. Aberystwyth");
 printf("\n\nWritten by Ian Pegler\n\n\n");
}


/* yyerror, as required by yacc to report syntax errors. It
   is also called from various other parts of the program. */

yyerror(s)
char *s;
{
 fprintf(stderr,"Error reading case file - %s\n",s);
 fprintf(stderr,"Line no %d",lines);
 exit(1);
}


/* display the introduction */

void displayintro()
{
struct intronode *pointer;
char *chptr;

 pointer = introroot;

 printf("\n\n\n");

 while (pointer != NULL)
   {
    chptr = (char *) pointer->text;	/* scan for newline and tab chars */
    while (*chptr != '\0')
      {
       if (*chptr == '\\')
         {
          chptr++;
          if (*chptr == 'n')
            {
             printf("\n");
            }
          else if (*chptr == 't')
            {
             printf("\t");
            }
          else
            {
             continue;
            }
         }
       else
         {
          printf("%c",*chptr);
         }
       chptr++;
      }
    printf("\n");
    pointer = pointer->introlink;
   }
}


/* check the local field definitions in each case */

void checklocalfielddefs()
{
struct casenode *caseptr;
struct fieldnode *fieldptr;
struct typenode *fielddef1,*fielddef2;
struct enumnode *enum1,*enum2;

 caseptr = caseroot;

 while (caseptr != NULL)
   {
    fielddef1 = caseptr->fielddefhead;
    while (fielddef1 != NULL)
       {

        /* check that fieldnames are not declared globally */

        fielddef2 = probdefnroot;
        while (fielddef2 != NULL)
           {
            if (strcmp(fielddef1->fieldname,fielddef2->fieldname) == 0)
              {
               fprintf(stderr,"Error - field %s is a global field in case %s\n",
                      fielddef1->fieldname,caseptr->name);
               exit(1);
              }
            fielddef2 = fielddef2->typelink;
           }

        /* check that fields are not doubly defined */

        fielddef2 = caseptr->fielddefhead;
        while (fielddef2 != NULL)
          {
           if (fielddef1 != fielddef2 &&
               strcmp(fielddef1->fieldname,fielddef2->fieldname) == 0)
             {
              fprintf(stderr,"Error - field %s is defined twice in case %s\n",
                     fielddef1->fieldname,caseptr->name);
              exit(1);
             }
           fielddef2 = fielddef2->typelink;
          }

        /* check that fieldnames are not used in solution part */

        fieldptr = caseptr->solnhead;
        while (fieldptr != NULL)
          {
           if (strcmp(fielddef1->fieldname,fieldptr->fieldname) == 0)
             {
              fprintf(stderr,"Error - locally defined field %s is used in solution\
 in case %s\n",
                     fielddef1->fieldname,caseptr->name);
              exit(1);
             }
           fieldptr = fieldptr-> fieldlink;
          }

        /* check for doubly defined enumeration symbols */

        if (fielddef1->fieldtype == ENUMTYPE)
          {
           enum1 = fielddef1->enumhead;
           while (enum1 != NULL)
             {
              enum2 = fielddef1->enumhead;
              while (enum2 != NULL)
                {
                 if (enum1 != enum2 &&
                     strcmp(enum1->evalue,enum2->evalue) == 0)
                   {
                    fprintf(stderr,"Error - symbol %s is defined twice in case %s\n",
                           enum1->evalue,caseptr->name);
                    exit(1);
                   }
                 enum2 = enum2->enumlink;
                }
              enum1 = enum1->enumlink;
             }
          }
        fielddef1 = fielddef1->typelink;
       }
    caseptr = caseptr->caselink;
   }
}


/* check the index block */

void checkindexblock()
{
struct indexnamenode *pointer;

 /* check that the type of the fields is ENUMTYPE */

 pointer = indexnameroot;

 while (pointer != NULL)
   {
    if (findfield(pointer->fieldname,probdefnroot) == NULL)
      {
       fprintf(stderr,"Error - Fieldname %s not defined in case definition\n",
               pointer->fieldname);
       exit(1);
      }
    checkfieldtype(pointer->fieldname, ENUMTYPE);
    pointer = pointer->indexnamelink;
   }
}


/* check repair rules attached to cases */

void checklocalreprules()
{
struct casenode *caseptr;

 caseptr = caseroot;

 while (caseptr != NULL)	/* for each case in the case list */
   {
    checkreprules(caseptr->rephead,caseptr);/* check the repair rules */
    caseptr = caseptr->caselink;
   }
}


/* check preprocessing rules for illegal presence of reselect */

void checkreselect()
{
struct repnode *ruleptr;
struct replistnode *actionptr;

 ruleptr = preprocessroot;

 while (ruleptr != NULL)
   {
    actionptr = ruleptr->replisthead;

    while (actionptr != NULL)
      {
       if (actionptr->reselect)
         {
          fprintf(stderr,"\nError - preprocessing rule %s contains reselect\n",
             ruleptr->rulename);
          exit(1);
         }

       actionptr = actionptr->replistlink;
      }

    ruleptr = ruleptr->replink;
   }
}


/* check repair rules */

void checkreprules(reprootptr,caseptr)
struct repnode *reprootptr;
struct casenode *caseptr;	/* pointer to case containing local repairs */
{
struct repnode *repptr;
struct replistnode *replistptr,*pointer;


 repptr = reprootptr;

 while (repptr != NULL)	/* check each repair rule */
   {
    checkrepname(repptr,reprootptr);	/* check for doubly used rulenames */

    checkcombtree(repptr->combtreeroot,caseptr);/* check combination tree */
    replistptr = repptr->replisthead;	/* check the action list */
    doublecheckreplist(replistptr);	/* check for doubly used fieldnames */

    pointer = replistptr;

    while (pointer != NULL)
      {
       if (pointer->undefined == FALSE &&
           pointer->print == FALSE &&
           pointer->restart == FALSE &&
           pointer->listop == FALSE &&
           pointer->reselect == FALSE &&
           pointer->extract == FALSE &&
           pointer->weightchange == FALSE &&
           (findfield(pointer->fieldname,probdefnroot) != NULL ||
            (caseptr != NULL &&
            findfield(pointer->fieldname,caseptr->fielddefhead) != NULL)))
         {
          localchecktype(pointer->fieldtype,pointer->fieldname,caseptr);
          if (pointer->fieldtype == ENUMTYPE)
            {	/* check that enumeration value is ok */
             localcheckenum(pointer->fieldname,pointer->strvalue,caseptr);
            }
         }
       else if (pointer->extract == TRUE)
         {
          if (!(abstract(pointer->abstraction)))
            {
             fprintf(stderr,"\nError - Abstract value required in extract operation\n");
             fprintf(stderr,"        Rule: %s\n",repptr->rulename);
             if (caseptr != NULL)
               {
                fprintf(stderr,"        Case: %s\n",caseptr->name);
               }
             exit(1);
            }
          if ((caseptr != NULL &&
               findfield(pointer->strvalue,caseptr->fielddefhead)) ||
              findfield(pointer->strvalue,probdefnroot))
            {
             localchecktype(LISTTYPE,pointer->strvalue,caseptr);
            }
          if ((caseptr != NULL &&
               findfield(pointer->fieldname,caseptr->fielddefhead)) ||
              findfield(pointer->fieldname,probdefnroot))
            {
             localchecktype(ENUMTYPE,pointer->fieldname,caseptr);
            }
         }
       pointer = pointer->replistlink;
      }

    repptr = repptr->replink;
   }
}


/* check for doubly used repair rule names */
void checkrepname(repptr,reprootptr)
struct repnode *repptr,*reprootptr;
{
struct repnode *pointer;

 pointer = reprootptr;

 while (pointer != NULL)	/* for each repair rule */
   {
    if (pointer != repptr &&	/* two different rules */
        strcmp(pointer->rulename,repptr->rulename) == 0)	/* same name */
      {
       fprintf(stderr,"Error - two repair rules with the same name: %s\n\n",
               pointer->rulename);
       exit(1);
      }
    pointer = pointer->replink;
   }
}


/* check the contents of a combination tree */

void checkcombtree(combtreeptr,caseptr)
struct combtreenode *combtreeptr;
struct casenode *caseptr;
{
 if (combtreeptr->optype == ITEM)
   {
    checkcombnode(combtreeptr->left,caseptr);
   }
 else if (combtreeptr->optype == NOT_OP)
   {
    checkcombtree(combtreeptr->left,caseptr);
   }
 else if (combtreeptr->optype == OR_OP ||
          combtreeptr->optype == AND_OP)
   {
    checkcombtree(combtreeptr->left,caseptr);
    checkcombtree(combtreeptr->right,caseptr);
   }
 else
   {
    fprintf(stderr,"Huh? - Unknown logical operator!\n");
    exit(1);
   }
}


/* check the field combination node */

void checkcombnode(pointer,caseptr)
struct combnode *pointer;
struct casenode *caseptr;	/* pointer to case if local repairs */
{
 /* type check if field is global */

 if (pointer->listop == FALSE &&
     pointer->fieldtype != NUMTYPE &&
    (findfield(pointer->fieldname,probdefnroot) != NULL ||
     (caseptr != NULL &&
     findfield(pointer->fieldname,caseptr->fielddefhead) != NULL)))
   {
    if (checkabstractstring(pointer,caseptr) == FALSE)
      {
       if (pointer->undefined == FALSE)
         {
          localchecktype(pointer->fieldtype,pointer->fieldname,caseptr);
         }
       if (pointer->undefined == FALSE &&
           pointer->fieldtype == ENUMTYPE)
         {	/* check that enumeration value is ok */
          localcheckenum(pointer->fieldname,pointer->strcontents,caseptr);
         }
      }
   }


}


/* return true if the combnode compares a string field against an
   abstract symbol */

int checkabstractstring(combitem,caseptr)
struct combnode *combitem;
struct casenode *caseptr;
{
struct typenode *fieldentry;

 if (combitem->fieldtype != ENUMTYPE)
   {
    return FALSE;
   }

 fieldentry = NULL;

 if (caseptr != NULL)
   {
    fieldentry = findfield(combitem->fieldname,caseptr->fielddefhead);
   }

 if (fieldentry == NULL)
   {
    fieldentry = findfield(combitem->fieldname,probdefnroot);
    if (fieldentry == NULL)
      {
       fprintf(stderr,"Error - field %s not defined, checking comparison\n",
          combitem->fieldname);
       exit(1);
      }
   }

 if (fieldentry->fieldtype != STRTYPE)
   {
    return FALSE;
   }

 return abstract(combitem->strcontents);
}


/* check that the field is defined */
void checkfieldexists(fieldname,defnroot)
char *fieldname;
struct typenode *defnroot;
{

 if (findfield(fieldname,defnroot) == NULL)
   {
    fprintf(stderr,"Error - Field %s is not defined\n",fieldname);
    exit(1);
   }
}


/* find the named field definition */
struct typenode *findfield(fieldname,defnroot)
struct typenode *defnroot;
buftype fieldname;
{
struct typenode *pointer;
 
 pointer = defnroot;
 
 while (pointer != NULL)
   {
    if (strcmp(pointer->fieldname,fieldname) == 0)	/* found field definition */
      {
       return pointer;
      }
    pointer = pointer->typelink;
   }
 
 return NULL;
}


/* return TRUE if named field exists in field list, FALSE otherwise */
int exists(fieldname,fieldlist)
char *fieldname;
struct fieldnode *fieldlist;
{
struct fieldnode *pointer;

 pointer = fieldlist;

 while (pointer != NULL)
   {
    if (strcmp(pointer->fieldname, fieldname) == 0)
       return TRUE;
    pointer = pointer->fieldlink;
   }

 return FALSE;
}


/* check replist for doubly used fieldnames */

void doublecheckreplist(replistptr)
struct replistnode *replistptr;
{
struct replistnode *pointer1,*pointer2;

 pointer1 = replistptr;
 while (pointer1 != NULL)
   {
    pointer2 = replistptr;
    while (pointer2 != NULL)
      {
       if (pointer1 != pointer2)
         {
          if (pointer1->restart == FALSE &&
              pointer2->restart == FALSE &&
              pointer1->print == FALSE &&
              pointer2->print == FALSE &&
              pointer1->reselect == FALSE &&
              pointer2->reselect == FALSE &&
              pointer1->listop == FALSE &&
              pointer2->listop == FALSE &&
              pointer1->weightchange == FALSE &&
              pointer2->weightchange == FALSE &&
              strcmp(pointer1->fieldname,pointer2->fieldname) == 0)
            {
             fprintf(stderr,"\nError - Fieldname %s used twice in repair rule\n",
                    pointer1->fieldname);
             exit(1);
            }
         }
       pointer2 = pointer2->replistlink;
      }
    pointer1 = pointer1->replistlink;
   }
}


void checkmodrules()	/* check modification rules */
{
struct modnode *modptr;
struct enumnode *pointer1, *pointer2;

 modptr = modroot;

 while (modptr != NULL)
   {
    if (modptr->fieldtype == NUMTYPE)
      {
       if (findfield(modptr->fieldname, probdefnroot) == NULL)
         {
          fprintf(stderr,"Error - Fieldname %s not defined in case definition\n",
                  modptr->fieldname);
          exit(1);
         }

       checkfieldtype(modptr->fieldname, modptr->fieldtype);

		/* do nothing more for number ranges */

       modptr = modptr->modlink;
       continue;
      }


	/* deal with abstraction rules */

    pointer1 = modptr->enumhead; /* check for doubly defined symbols */
    while (pointer1 != NULL)
      {

       /* check that abstraction symbol is not in list */
       if (strcmp(modptr->abstraction,pointer1->evalue) == 0)
         {
          fprintf(stderr,"Error - abstraction value %s included in similarity list\n",
                    modptr->abstraction);
          exit(1);
         }

       pointer2 = modptr->enumhead;
       while (pointer2 != NULL)
         {
          if (pointer1 != pointer2)
            { 
             if (strcmp(pointer1->evalue, pointer2->evalue) == 0)
               {
                fprintf(stderr,"Error - similarity list contains %s defined twice\n",
                                                     pointer1->evalue);
                exit(1);
               }
            } 
          pointer2 = pointer2->enumlink;
         }
       pointer1 = pointer1->enumlink;
      }

    modptr = modptr->modlink;
   }

 checkabstraction();	/* check for doubly defined abstraction symbols */
}
 

/* check for doubly defined abstraction symbols */

void checkabstraction()
{
struct modnode *pointer1,*pointer2;

 pointer1 = modroot;

 while (pointer1 != NULL)
   {
    if (pointer1->fieldtype == ENUMTYPE)
      {
       pointer2 = modroot;
       while (pointer2 != NULL)
         {
          if (pointer2 != pointer1 &&
              pointer2->fieldtype == ENUMTYPE)
            {
             if (strcmp(pointer1->abstraction,pointer2->abstraction) == 0)
               {
                fprintf(stderr,"Error - doubly defined abstraction symbol %s\n",
                        pointer1->abstraction);
                exit(1);
               }
            }
          pointer2 = pointer2->modlink;
         }
      }
    pointer1 = pointer1->modlink;
   }
}


/* check that named field is defined and has the specified type */

void checkfieldtype(fieldname,fieldtype)
char *fieldname;
int fieldtype;
{
struct typenode *fieldptr;

 fieldptr = findfield(fieldname,probdefnroot);
 if (fieldptr == NULL)
   {
    fprintf(stderr,"Error - fieldname %s not defined\n",fieldname);
    exit(1);
   }

 if (fieldptr->fieldtype != fieldtype)
   {
    fprintf(stderr,"Error - Type mismatch error on field %s\n",fieldname);
    exit(1);
   }
}


void doublecheckenums(defnroot) /* check for doubly defined enumeration symbols */
struct typenode *defnroot;
{
struct typenode *typeptr;
struct enumnode *pointer1, *pointer2;

 if (defnroot == NULL)
   {
    fprintf(stderr,"Huh? NULL definition list!");
    exit(1);
   }

 typeptr = defnroot;

 while (typeptr != NULL)
   {
    if (typeptr->fieldtype == ENUMTYPE)
      {
       pointer1 = typeptr->enumhead;

       /* ensure enumeration list is long enough */
       if (enumlistlength(pointer1) < 2)
         {
          fprintf(stderr,"Error - field %s enumeration list needs at least two values\n",
                        typeptr->fieldname);
          exit(1);
         }

       while (pointer1 != NULL)
         {
          /* check that fieldname is not in list */
          if (strcmp(typeptr->fieldname,pointer1->evalue) == 0)
            {
             fprintf(stderr,"Error - fieldname %s in enumeration list\n",typeptr->fieldname);
             exit(1);
            }

          pointer2 = typeptr->enumhead;
          while (pointer2 != NULL)
            {
             if (pointer1 != pointer2)
               {
                if (strcmp(pointer1->evalue, pointer2->evalue) == 0)
                  {
                   fprintf(stderr,"Error - enumeration list contains %s defined twice\n",
							pointer1->evalue);
                   exit(1);
                  }
               }
             pointer2 = pointer2->enumlink;
            }
          pointer1 = pointer1->enumlink;
         }
      }
    typeptr = typeptr->typelink;
   }

}


/* returns the length of an enumeration list */

int enumlistlength(enumptr)
struct enumnode *enumptr;
{
struct enumnode *pointer;
int length;

 length = 0;
 pointer = enumptr;

 while (pointer != NULL)
   {
    length++;
    pointer = pointer->enumlink;
   }

 return length;
}


void doublecheckcases() /* check for doubly defined case names */
{
struct casenode *pointer1, *pointer2;

 pointer1 = caseroot;

 while (pointer1 != NULL)
   {
    pointer2 = caseroot;
    while (pointer2 != NULL)
      {
       if (pointer1 != pointer2)
         {
          if (strcmp(pointer1->name, pointer2->name) == 0)
            {
             fprintf(stderr,"Error - case name %s is not unique\n",pointer1->name);
             exit(1);
            }
         }
       pointer2 = pointer2->caselink;
      }
    pointer1 = pointer1->caselink;
   }

}


void multiplecheckfields() /* check each case instance for multiple uses of a field */
{
struct casenode *caseptr;

 caseptr = caseroot;

 while (caseptr != NULL)
   {
    checkmultcase(caseptr);
    caseptr = caseptr->caselink;
   }
}

void checkmultcase(caseptr)
struct casenode *caseptr;
{

 checkfieldlist(caseptr->fieldhead, caseptr->name);
 checkfieldlist(caseptr->solnhead, caseptr->name);
}


void checkfieldlist(fieldptr, name) /* check a field list for multiple fields */
struct fieldnode *fieldptr;
char *name;
{
struct fieldnode *pointer1, *pointer2;

 pointer1 = fieldptr;

 while (pointer1 != NULL)
   {
    pointer2 = fieldptr;
    while (pointer2 != NULL)
      {
       if (pointer1 != pointer2)
         {
          if (strcmp(pointer1->fieldname, pointer2->fieldname) == 0)
            {
             fprintf(stderr,"Error - field %s used twice in case %s\n",
                                                pointer1->fieldname, name);
             exit(1);
            }
         }
       pointer2 = pointer2->fieldlink;
      }
    pointer1 = pointer1->fieldlink;
   }

}

    
void doublecheckfields(defnroot) /* check that each field definition is unique */
struct typenode *defnroot;
{
struct typenode *pointer1, *pointer2;

 pointer1 = defnroot;

 while (pointer1 != NULL)
   {
    pointer2 = defnroot;
    while (pointer2 != NULL)
      {
       if (pointer1 != pointer2)
         {
          if (strcmp(pointer1->fieldname, pointer2->fieldname) == 0)
            {
             fprintf(stderr,"Error - Fieldname %s defined twice\n", pointer1->fieldname);
             exit(1);
            }
         }
       pointer2 = pointer2->typelink;
      }
    pointer1 = pointer1->typelink;
   }
}


void typecheckfields() /* check the types of all the fields */
{
struct casenode *thiscase;

 thiscase = caseroot;
 if (caseroot == NULL)
   {
    fprintf(stderr,"Huh? NULL case list!");
    exit(1);
   }

 while (thiscase != NULL)
   {
    checkcase(thiscase);
    thiscase = thiscase->caselink;
   }
}

void checkcase(thiscase) /* check the types of a case */
struct casenode *thiscase;
{
struct fieldnode *thisfield;

 thisfield = thiscase->fieldhead;

 if (thisfield == NULL)
   {
    fprintf(stderr,"Huh? NULL field list!");
    exit(1);
   }

 while (thisfield != NULL)
   {
    checkfield(thisfield, thiscase);
    thisfield = thisfield->fieldlink;
   }

}

void checkfield(thisfield, thiscase) /* check the type of a field */
struct fieldnode *thisfield;
struct casenode *thiscase;
{
 localchecktype(thisfield->fieldtype, thisfield->fieldname, thiscase);

 if (thisfield->fieldtype == ENUMTYPE)
   {
    localcheckenum(thisfield->fieldname, thisfield->strcontents, thiscase);
   }
}


void checktype(typespec, ident, nodeptr) /* check the type of the identifier */
int typespec;
buftype ident;
struct typenode *nodeptr;
{

 if (nodeptr == NULL) /* at the end of the field definitions */
   {
    /* anonymous field */
    return;
   }

 if (strcmp(nodeptr->fieldname, ident) == 0) /* found field name */
   {
    if (nodeptr->fieldtype == typespec) /* types match */
       return;
    else
      {
       fprintf(stderr,"Error - Type mismatch error on field %s\n",ident);
       exit(1);
      }

   }

 checktype(typespec, ident, nodeptr->typelink);
}


/* check type of field in local then global field definitions */

void localchecktype(fieldtype,fieldname,caseptr)
int fieldtype;
buftype fieldname;
struct casenode *caseptr;
{
struct typenode *fielddef;

 if (caseptr != NULL)
   {
    fielddef = caseptr->fielddefhead;

    while (fielddef != NULL)
      {
       if (strcmp(fieldname,fielddef->fieldname) == 0)
         {
          if (fielddef->fieldtype == fieldtype)
            {
             return;	/* types match */
            }
          else
            {
             fprintf(stderr,"Error - type mismatch on local field %s\n\
in case %s\n",fieldname,caseptr->name);
             exit(1);
            }
         }
       fielddef = fielddef->typelink;
      }
   }

 /* no match with local definitions so check global ones */

 checktype(fieldtype,fieldname,probdefnroot);
}


void localcheckenum(fieldname,evalue,caseptr)
buftype fieldname;
buftype evalue;
struct casenode *caseptr;
{
struct enumnode *head;

 if (abstract(evalue))
   {
    return;
   }

 if (caseptr != NULL)
   {
    head = searchtypelist(fieldname,caseptr->fielddefhead);
   }
 else
   {
    head = NULL;
   }

 if (head == NULL)
   {
    head = searchtypelist(fieldname, probdefnroot);
   }

 if (head == NULL)
   {
    /* anonymous field, so do nothing */
    return;
   }

 searchenodes(head, evalue);
}


void checkenum(fieldname, evalue) /* check enumeration value */

buftype fieldname;
buftype evalue;
{
struct enumnode *head;

 if (abstract(evalue))
   {
    return;
   }

 head = searchtypelist(fieldname, probdefnroot);
 if (head == NULL)
   {
    fprintf(stderr,"Error - Fieldsname: %s does not exist\n",fieldname);
    exit(1);
   }
 searchenodes(head, evalue);
}


struct enumnode *searchtypelist(fieldname, nodeptr) /* search for enumeration definition */

buftype fieldname;
struct typenode *nodeptr;
{
struct typenode *pointer;

 pointer = nodeptr;

 while (pointer != NULL)
   {

    if (strcmp(pointer->fieldname, fieldname) == 0) /* found field name */
      {
       return pointer->enumhead; /* return pointer to head of enumeration list*/
      }
    else
      {
       pointer = pointer->typelink;
      }
   }
 return NULL;
}


void searchenodes(enumptr, evalue) /* search for enumeration string */
struct enumnode *enumptr;
buftype evalue;
{
struct enumnode *pointer;

 pointer = enumptr;

 while (pointer != NULL)
   {
    if (strcmp(pointer->evalue, evalue) == 0)
      {
       return; /* symbol exists so finish */
      }
    else
      {
       pointer = pointer->enumlink;
      }
   }
 fprintf(stderr,"Error - Enumeration symbol %s not defined\n",evalue);
 exit(1);
}


/* create a list operation node for a head or tail operation */

struct listopnode *createlistopnode(operation,listop)
int operation;
struct listopnode *listop;
{
struct listopnode *pointer;

 pointer = (struct listopnode *) calloc(1, sizeof(struct listopnode));
				/* allocate space for listopnode */

 pointer->optype = operation;
 pointer->listoplink = listop;
 return pointer;
}


/* create a list operation nnode to hold a field name */

struct listopnode *createlistopfieldnode(fieldname)
buftype *fieldname;
{
struct listopnode *pointer;

 pointer = (struct listopnode *) calloc(1, sizeof(struct listopnode));
				/* allocate space for listopnode */

 pointer->optype = FIELDNAME;
 pointer->fieldname = fieldname;
 pointer->listoplink = NULL;
 return pointer;
}


/* create an introduction text node */

struct intronode *createintronode(text)
buftype *text;
{
struct intronode *pointer;

 pointer = (struct intronode *) calloc(1, sizeof(struct intronode));
		/* allocate space for intronode */

 pointer->text = text;
 pointer->introlink = NULL;
 return pointer;
}


/* create an index specification node */

struct indexspecnode *createindexspecnode(fieldname,value)
buftype *fieldname;
buftype *value;
{
struct indexspecnode *pointer;

 pointer = (struct indexspecnode *) calloc(1, sizeof(struct indexspecnode));
		/* allocate space for indexspecnode */

 strcpy(pointer->fieldname,fieldname);
 strcpy(pointer->value,value);
 pointer->indexspeclink = NULL;
 return pointer;
}


/* create an index name node */

struct indexnamenode *createindexnamenode(fieldname)
buftype *fieldname;
{
struct indexnamenode *pointer;

 pointer = (struct indexnamenode *) calloc(1, sizeof(struct indexnamenode));
		/* allocate space for indexname node */

 strcpy(pointer->fieldname, fieldname);
 pointer->indexhead = NULL;
 pointer->indexnamelink = NULL;
 return pointer;
}


/* create an index node */

struct indexnode *createindexnode(evalue)
buftype *evalue;
{
struct indexnode *pointer;

 pointer = (struct indexnode *) calloc(1, sizeof(struct indexnode));
		/* allocate space for indexnode */
 strcpy(pointer->evalue,evalue);
 pointer->caseptrhead = NULL;
 pointer->indexlink = NULL;
 return pointer;
}

/* create a case pointer node */

struct caseptrnode *createcaseptrnode(caseptr)
struct casenode *caseptr;
{
struct caseptrnode *pointer;

 pointer = (struct caseptrnode *) calloc(1, sizeof(struct caseptrnode));
		/* allocate space for caseptrnode */
 pointer->caseind = caseptr;
 pointer->caseptrlink = NULL;
 return pointer;
}


/* create a combination tree node */

struct combtreenode *createcombtreenode(operation,leftchild,rightchild)
int operation;
struct combtreenode *leftchild,*rightchild;
{
struct combtreenode *pointer;

 pointer = (struct combtreenode *) calloc(1, sizeof(struct combtreenode));
		/* allocate space for combtreenode */

 pointer->optype = operation;
 pointer->left = leftchild;
 pointer->right = rightchild;
 return pointer;
}


/* create an arithnode to hold an addition operation */

struct arithnode *createplusnode(operand1,operand2)
struct arithnode *operand1,*operand2;
{
struct arithnode *pointer;

 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));
		/* allocate space for arithnode */

 pointer->contentstype = PLUS;
 pointer->left = operand1;
 pointer->right = operand2;
 return pointer;
}


/* create an arithnode to hold a subtraction operation */

struct arithnode *createminusnode(operand1,operand2)
struct arithnode *operand1,*operand2; 
{
struct arithnode *pointer; 
 
 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode)); 
		/* allocate space for arithnode */ 
 
 pointer->contentstype = MINUS;
 pointer->left = operand1;
 pointer->right = operand2; 
 return pointer; 
}


/* create an arithnode to hold a multiplication operation */

struct arithnode *createmultiplynode(operand1,operand2) 
struct arithnode *operand1,*operand2;  
{ 
struct arithnode *pointer;

 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));  
		/* allocate space for arithnode */  
 
 pointer->contentstype = MULTIPLY; 
 pointer->left = operand1;    
 pointer->right = operand2;  
 return pointer;  
}


/* create an arithnode to hold a division operation */

struct arithnode *createdividenode(operand1,operand2) 
struct arithnode *operand1,*operand2;  
{ 
struct arithnode *pointer;  
  
 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));  
		/* allocate space for arithnode */  
  
 pointer->contentstype = DIVIDE; 
 pointer->left = operand1;    
 pointer->right = operand2;  
 return pointer;  
}


/* create an arithnode to hold a power operation */

struct arithnode *createpowernode(operand1,operand2) 
struct arithnode *operand1,*operand2;  
{ 
struct arithnode *pointer;  
 
 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));  
		/* allocate space for arithnode */  
  
 pointer->contentstype = POWER; 
 pointer->left = operand1;    
 pointer->right = operand2;  
 return pointer;  
}


/* create an arithnode to hold a unary minus operation */

struct arithnode *createunaryminusnode(operand)
struct arithnode *operand;
{
struct arithnode *pointer;

 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));
		/* allocate space for arithnode */

 pointer->contentstype = UMINUS;
 pointer->left = operand;
 return pointer;
}


/* create an arithnode to hold a number */

struct arithnode *createnumbernode(numvalue)
float numvalue;
{
struct arithnode *pointer;

 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));
		/* allocate space for arithnode */

 pointer->contentstype = NUMVALUE;
 pointer->numcontents = numvalue;
 return pointer;
}


/* create an arithnode to hold a fieldname */

struct arithnode *createfieldvaluenode(fieldname)
buftype *fieldname;
{
struct arithnode *pointer;

 pointer = (struct arithnode *) calloc(1, sizeof(struct arithnode));
		/* allocate space for arithnode */

 pointer->contentstype = FIELDNAME;
 strcpy(pointer->fieldname,fieldname);
 return pointer;
}


/* create a repnode */

struct repnode *createrepnode(rulename,combtreeptr,replistptr)
buftype *rulename;
struct combtreenode *combtreeptr;
struct replistnode *replistptr;
{
struct repnode *pointer;

 pointer = (struct repnode *) calloc(1, sizeof(struct repnode));
				/* allocate space for repnode */

 strcpy(pointer->rulename,rulename);
 pointer->combtreeroot = combtreeptr;
 pointer->replisthead = replistptr;
 pointer->replink = NULL;
 return pointer;
}


/* create a combnode to hold an enumeration value */

struct combnode *createcombenumnode(fieldname, value)
buftype *fieldname,*value;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
				/* allocate space for combnode */

 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = ENUMTYPE;
 pointer->undefined = FALSE;
 pointer->listop = FALSE;
 strcpy(pointer->strcontents, value);
 return pointer;
}


/* create a combnode to hold a comparison */

struct combnode *createcombcompnode(left, operator, right)

struct arithnode *left, *right;
int operator;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
				/* allocate space for combnode */

 pointer->fieldtype = NUMTYPE;
 pointer->undefined = FALSE;
 pointer->listop = FALSE;
 pointer->left = left;
 pointer->right = right;
 pointer->operator = operator;
 return pointer;
}


/* create a combnode to hold a string */

struct combnode *createcombstringnode(fieldname, value)
buftype *fieldname,*value;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
                                /* allocate space for combnode */
 
 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = STRTYPE;
 pointer->undefined = FALSE;
 pointer->listop = FALSE;
 strcpy(pointer->strcontents, value);
 return pointer;
}


/* create a combnode to hold a list */

struct combnode *createcomblistnode(fieldname, listptr)
buftype *fieldname;
struct listnode *listptr;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
				/* allocate space for combnode */

 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = LISTTYPE;
 pointer->undefined = FALSE;
 pointer->listop = FALSE;
 pointer->listvalue = listptr;
 return pointer;
}


/* create a combnode to hold a list operation on an */
/* enumeration value */

struct combnode *createcomblistopenumnode(listop, value)
buftype *value;
struct listopnode *listop;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
				/* allocate space for combnode */

 pointer->fieldtype = ENUMTYPE;
 pointer->undefined = FALSE;
 pointer->listop = TRUE;
 strcpy(pointer->strcontents,value);
 pointer->listophead = listop;
 return pointer;
}


/* create a combnode to hold a list operation on an */
/* expression */

struct combnode *createcomblistopexpnode(listop, operator, tree)
int operator;
struct arithnode *tree;
struct listopnode *listop;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
                                /* allocate space for combnode */

 pointer->fieldtype = NUMTYPE;
 pointer->undefined = FALSE;
 pointer->listop = TRUE;
 pointer->operator = operator;
 pointer->right = tree;
 pointer->listophead = listop;
 return pointer;
}


/* create a combnode to hold a list operation on a */
/* string value */

struct combnode *createcomblistopstringnode(listop, value)
buftype *value;
struct listopnode *listop;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
                                /* allocate space for combnode */

 pointer->fieldtype = STRTYPE;
 pointer->undefined = FALSE;
 pointer->listop = TRUE;
 strcpy(pointer->strcontents,value);
 pointer->listophead = listop;
 return pointer;
}


/* create a combnode to hold a list operation on a */
/* list value */

struct combnode *createcomblistoplistnode(listop, listvalue)
struct listnode *listvalue;
struct listopnode *listop;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
                                /* allocate space for combnode */

 pointer->fieldtype = LISTTYPE;
 pointer->undefined = FALSE;
 pointer->listop = TRUE;
 pointer->listvalue = listvalue;
 pointer->listophead = listop;
 return pointer;
}


/* create a combnode to hold UNDEFINED */

struct combnode *createcombundefnode(fieldname)
buftype *fieldname;
{
struct combnode *pointer;

 pointer = (struct combnode *) calloc(1, sizeof(struct combnode));
				/* allocate space for combnode */

 strcpy(pointer->fieldname, fieldname);
 pointer->undefined = TRUE;
 pointer->listop = FALSE;
 return pointer;
}


/* create a repair action node to hold an enumerated type */
 
struct replistnode *createreplistenumnode(fieldname, value)
buftype *fieldname, *value;
{ 
struct replistnode *pointer;
 
 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
                                /* allocate space for replistnode */
 
 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = ENUMTYPE;
 pointer->undefined = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->restart = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->strvalue = value;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action node to hold an expression tree */
 
struct replistnode *createreplistexpnode(fieldname, exptree)
buftype *fieldname;
struct arithnode *exptree;
{ 
struct replistnode *pointer; 
 
 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode)); 
                                /* allocate space for replistnode */ 
 
 strcpy(pointer->fieldname, fieldname); 
 pointer->fieldtype = NUMTYPE; 
 pointer->undefined = FALSE; 
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->exptree = exptree;
 pointer->replistlink = NULL; 
 return pointer;
}


/* create a repair action node to hold a list */

struct replistnode *createreplistlistnode(fieldname, listptr)
buftype *fieldname;
struct listnode *listptr;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = LISTTYPE;
 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->tolist = listptr;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action node to change a */
/* list-operation specified list element to */
/* an enumeration value */

struct replistnode *createreplistlistopenumnode(listop, value)
buftype *value;
struct listopnode *listop;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 pointer->fieldtype = ENUMTYPE;
 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = TRUE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->strvalue = value;
 pointer->listophead = listop;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action node to change a */
/* list-operation specified list element to */
/* an expression */

struct replistnode *createreplistlistopexpnode(listop, tree)
struct arithnode *tree;
struct listopnode *listop;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
                                /* allocate space for replistnode */

 pointer->fieldtype = NUMTYPE;
 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = TRUE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->exptree = tree;
 pointer->listophead = listop;
 pointer->replistlink = NULL;
 return pointer;
}

/* create a repair action node to change a */
/* list-operation specified list element to */
/* a string */

struct replistnode *createreplistlistopstringnode(listop, value)
buftype *value;
struct listopnode *listop;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
                                /* allocate space for replistnode */

 pointer->fieldtype = STRTYPE;
 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = TRUE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->strvalue = value;
 pointer->listophead = listop;
 pointer->replistlink = NULL;
 return pointer;
}

/* create a repair action node to change a */
/* list-operation specified list element to */
/* a list */

struct replistnode *createreplistlistoplistnode(listop, listvalue)
struct listnode *listvalue;
struct listopnode *listop;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
                                /* allocate space for replistnode */

 pointer->fieldtype = LISTTYPE;
 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = TRUE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->tolist = listvalue;
 pointer->listophead = listop;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action node to hold a string */

struct replistnode *createrepliststringnode(fieldname, strvalue)
buftype *fieldname,*strvalue;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
                                /* allocate space for replistnode */

 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = STRTYPE;
 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->strvalue = strvalue;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action to undefine a node */

struct replistnode *createreplistundefnode(fieldname)
buftype *fieldname;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replist node */

 strcpy(pointer->fieldname, fieldname);
 pointer->undefined = TRUE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action to restart the repair rules */

struct replistnode *createreplistrepnode()
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 pointer->undefined = FALSE;
 pointer->restart = TRUE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action to print a string */

struct replistnode *createreplistprintnode(listptr)
struct listnode *listptr;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = TRUE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->tolist = listptr;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action to reselect a case for repair */

struct replistnode *createreplistresnode()
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = TRUE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = FALSE;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action for an extract operation */

struct replistnode *createreplistextractnode(fieldname,abstraction,listname)
buftype *fieldname,*abstraction,*listname;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = TRUE;
 pointer->weightchange = FALSE;
 strcpy(pointer->fieldname,fieldname);
 pointer->abstraction = abstraction;
 pointer->strvalue = listname;
 pointer->fieldtype = ENUMTYPE;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a repair action for a weight change operation */

struct replistnode *createreplistweightnode(fieldname,newweight)
buftype *fieldname;
float newweight;
{
struct replistnode *pointer;

 pointer = (struct replistnode *) calloc(1, sizeof(struct replistnode));
				/* allocate space for replistnode */

 pointer->undefined = FALSE;
 pointer->restart = FALSE;
 pointer->print = FALSE;
 pointer->reselect = FALSE;
 pointer->listop = FALSE;
 pointer->extract = FALSE;
 pointer->weightchange = TRUE;
 strcpy(pointer->fieldname,fieldname);
 pointer->weight = newweight;
 pointer->replistlink = NULL;
 return pointer;
}


/* create a node to hold the text of a string */

char *createstringnode()
{
char *pointer;

 pointer = (char *) calloc(1, sizeof(buftype));

 return pointer;
}


/* create a buftype, needed for type compatibility */

buftype *createbuftypenode()
{
buftype *pointer;

 pointer = (buftype *) calloc(1, sizeof(buftype));

 return pointer;
}


/* create field node to hold a string value */

struct fieldnode *createstrfield(fieldname, contents)
buftype fieldname;
buftype *contents;
{

struct fieldnode *pointer;

 pointer = (struct fieldnode *) calloc(1, sizeof(struct fieldnode));
				/* allocate space for node */

 pointer->strcontents = contents;
 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = STRTYPE;
 pointer->fieldlink = NULL;
 return pointer;
}


/* create field node to hold a number */

struct fieldnode *createnumfield(fieldname, contents)
buftype fieldname;
float contents;
{
 
struct fieldnode *pointer;
 
 pointer = (struct fieldnode *) calloc(1, sizeof(struct fieldnode));
                                /* allocate space for node */
 
 pointer->numcontents = contents;
 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = NUMTYPE; 
 pointer->fieldlink = NULL; 
 return pointer; 
}


/* create field node to hold a list */
 
struct fieldnode *createlistfield(fieldname, contents)
buftype fieldname;
struct listnode *contents;
{
 
struct fieldnode *pointer;
 
 pointer = (struct fieldnode *) calloc(1, sizeof(struct fieldnode));
                                /* allocate space for node */
 
 pointer->listcontents = contents;
 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = LISTTYPE;
 pointer->fieldlink = NULL;
 return pointer;
}


/* create field node to hold an enumeration value */

struct fieldnode *createenumfield(fieldname, contents)
buftype fieldname;
buftype *contents;
{
 
struct fieldnode *pointer;
 
 pointer = (struct fieldnode *) calloc(1, sizeof(struct fieldnode));
                                /* allocate space for node */
 
 pointer->strcontents = contents;
 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = ENUMTYPE; 
 pointer->fieldlink = NULL; 
 return pointer; 
}


/* create a case node, fieldhead points to the field list */

struct casenode *createcasenode(casename)
buftype casename;
{

struct casenode *pointer;

 pointer = (struct casenode *) calloc(1, sizeof(struct casenode));
				/* allocate space for node */

 strcpy(pointer->name, casename);
 pointer->fieldhead = NULL;
 pointer->solnhead = NULL;
 pointer->rephead = NULL;
 pointer->caselink = NULL;
 return pointer;
}

/* create enumeration symbol node. Defines a value that an enumeration
   symbol may take */

struct enumnode *createenumnode(evalue)
buftype evalue;
{

struct enumnode *pointer;

 pointer = (struct enumnode *) calloc(1, sizeof(struct enumnode));
				/* allocate space for node */

 strcpy(pointer->evalue, evalue);
 pointer->enumlink = NULL;
 return pointer;
}


/* create a node to define the type of a field. If the type is
   an enumeration, then enumhead points to the list of values
   that the type may take */

struct typenode *createtypenode(fieldname, fieldtype, weight, prompt)
buftype *fieldname;
int fieldtype;
float weight;
struct listnode *prompt;
{

struct typenode *pointer;

 pointer = (struct typenode *) calloc(1, sizeof(struct typenode));
				/* allocate space for node */

 strcpy(pointer->fieldname, fieldname);
 pointer->fieldtype = fieldtype;
 pointer->enumhead = NULL;
 pointer->typelink = NULL;
 pointer->weight = weight;
 pointer->oldweight = weight;
 pointer->prompt = prompt;
 return pointer;
}


/* create list node to hold an enumeration (symbol) value */

struct listnode *createlistenum(value)
buftype *value;
{

struct listnode *pointer;

 pointer = (struct listnode *) calloc(1, sizeof(struct listnode));
				/* allocate space for list node */

 pointer->strcontents = value;
 pointer->listtype = ENUMTYPE;
 pointer->listlink = NULL;
 return pointer;
}


/* create list node to hold a real value */

struct listnode *createlistnum(value)
float value;
{

struct listnode *pointer;

 pointer = (struct listnode *) calloc(1, sizeof(struct listnode));
                                /* allocate space for list node */

 pointer->numcontents = value;
 pointer->listtype = NUMTYPE;
 pointer->listlink = NULL;
 return pointer;
}


/* create list node to hold a string value */

struct listnode *createliststring(value)
buftype *value;
{

struct listnode *pointer;

 pointer = (struct listnode *) calloc(1, sizeof(struct listnode));
                                /* allocate space for list node */

 pointer->strcontents = value;
 pointer->listtype = STRTYPE;
 pointer->listlink = NULL;
 return pointer;
}


/* create list node to hold a list */

struct listnode *createlistlist(value)
struct listnode *value;
{

struct listnode *pointer;

 pointer = (struct listnode *) calloc(1, sizeof(struct listnode));
                                /* allocate space for list node */

 pointer->listitem = value;
 pointer->listtype = LISTTYPE;
 pointer->listlink = NULL;
 return pointer;
}


/* create a modification node to hold an enumeration list */

struct modnode *createmodenumnode(abstraction,listpointer)
buftype *abstraction;
struct enumnode *listpointer;
{
struct modnode *pointer;

 pointer = (struct modnode *) calloc(1, sizeof(struct modnode));
				/* allocate space for modification node */

 strcpy(pointer->abstraction, abstraction);
 pointer->fieldtype = ENUMTYPE;
 pointer->enumhead = listpointer;
 pointer->modlink = NULL;
 return pointer;
}


/* create a modification node to hold a number range */

struct modnode *createmodrangenode(name,start,end)
char *name;
float start,end;
{
struct modnode *pointer;

 pointer = (struct modnode *) calloc(1, sizeof(struct modnode));
				/* allocate space for modification node */

 strcpy(pointer->fieldname,name);
 pointer->fieldtype = NUMTYPE;
 pointer->start = start;
 pointer->end = end;
 pointer->modlink = NULL;
 return pointer;
}


/* display the contents of a case */
/* local repair rules are not displayed */

void displaycase(thiscase)
struct casenode *thiscase;
{
struct fieldnode *thisfield;

 if (thiscase == NULL)
   {
    fprintf(fp,"No cases!\n");
    return;
   }

 fprintf(fp,"\n\ncase instance %s is\n\n",thiscase->name);
 thisfield = thiscase->fieldhead;
 while (thisfield != NULL)
   {
    fprintf(fp,"\t%s = ",thisfield->fieldname);
    displayfield(thisfield,thiscase); 
    fprintf(fp,";\n");
    thisfield = thisfield->fieldlink;
   }
 fprintf(fp,"\nsolution is\n\n");
 thisfield = thiscase->solnhead;
 while (thisfield != NULL)
   {
    fprintf(fp,"\t%s = ",thisfield->fieldname);
    displayfield(thisfield,thiscase);
    fprintf(fp,";\n");
    thisfield = thisfield->fieldlink;
   }
 fprintf(fp,"\nend;\n");
}


/* display the contents of a field */

void displayfield(thisfield,thiscase)
struct casenode *thiscase;
struct fieldnode *thisfield;
{
 if (thisfield->fieldtype == STRTYPE)
   {
    fprintf(fp,"\'%s\'",thisfield->strcontents);
   }
 else if (thisfield->fieldtype == NUMTYPE)
   {
    fprintf(fp,"%.5f",thisfield->numcontents);
   }
 else if (thisfield->fieldtype == LISTTYPE)
   {
    displaylist(thisfield->listcontents,thiscase);
   }
 else if (thisfield->fieldtype == ENUMTYPE)
   {
    fprintf(fp,"%s",thisfield->strcontents);
   }
 else
   {
    fprintf(fp,"NONSENSE");
   }
}


/* display the contents of a field, expand contents */

void displayfieldexpand(thisfield,thiscase)
struct casenode *thiscase;
struct fieldnode *thisfield;
{
 if (thisfield->fieldtype == STRTYPE)
   {
    fprintf(fp,"%s",thisfield->strcontents);
   }
 else if (thisfield->fieldtype == NUMTYPE)
   {
    fprintf(fp,"%.5f",thisfield->numcontents);
   }
 else if (thisfield->fieldtype == LISTTYPE)
   {
    displaylist(thisfield->listcontents,thiscase);
   }
 else if (thisfield->fieldtype == ENUMTYPE)
   {
    fprintf(fp,"%s",thisfield->strcontents);
   }
 else
   {
    fprintf(fp,"NONSENSE");
   }
}
/* display the contents of a list */

void displaylist(listptr,thiscase)
struct listnode *listptr;
struct casenode *thiscase;
{
 if (listexpansion == TRUE)
   {
    fprintf(fp,"\n");
   }
 else
   {
    fprintf(fp,"[ ");
   }
 displaylistitems(listptr,thiscase);
 if (listexpansion == FALSE)
   {
    fprintf(fp," ]");
   }
}

void displaylistitems(listptr,thiscase)
struct listnode *listptr;
struct casenode *thiscase;
{
struct fieldnode *valuefield;

 if (listptr != NULL)
   {
    if (listptr->listtype == STRTYPE)
      {
       if (listexpansion == TRUE)
         {
          fprintf(fp,"%s", listptr->strcontents);
         }
       else
         {
          fprintf(fp,"\'%s\'", listptr->strcontents);
         }
      }
    else if (listptr->listtype == ENUMTYPE)
      {
       if (listexpansion == TRUE)
         {
          valuefield = searchfield(listptr->strcontents,thiscase->fieldhead);
          if (valuefield == NULL)
            {
             valuefield = searchfield(listptr->strcontents,thiscase->solnhead);
            }

          if (valuefield != NULL)	/* do list expansion */
            {
             displayfieldexpand(valuefield,thiscase);
            }
          else
            {
             fprintf(fp,"%s",listptr->strcontents);
            }
         }
       else
         {
          fprintf(fp,"%s",listptr->strcontents);
         }
      }
    else if (listptr->listtype == NUMTYPE)
      {
       fprintf(fp,"%.5f",listptr->numcontents);
      }
    else if (listptr->listtype == LISTTYPE)
      {
       displaylist(listptr->listitem,thiscase);
      }
    else
      {
       fprintf(fp,"NONSENSE");
      }

    if (listptr->listlink != NULL)
      {
       fprintf(fp," ");
       displaylistitems(listptr->listlink,thiscase);
      }

   }
}

#include "casematching.c"
#include "rulerepair.c"
#include "evaluate.c"
#include "buildindexes.c"
#include "menu.c"
#include "addtocasebase.c"
#include "cleanup.c"

int yyexca[] ={
-1, 1,
	0, -1,
	-2, 0,
	};
# define YYNPROD 111
# define YYLAST 418
int yyact[]={

   150,   169,   146,   129,   209,   156,   170,   171,   171,   158,
   170,   171,   146,   129,   126,   191,   158,   158,   282,   264,
   156,   260,   129,    92,   129,    89,    95,    62,    60,    58,
    57,   159,    36,   130,    84,   237,   249,   235,   159,   159,
    75,   122,    50,   277,   238,   228,   253,   252,   129,   250,
   142,   132,   113,   202,    69,   246,   203,    80,   129,    40,
   129,    29,   248,    66,   232,    28,    82,   100,    19,    48,
   251,   206,   184,   182,   185,   204,   205,    13,   198,   108,
   268,   175,   164,   134,   116,   112,   101,    76,    68,    63,
    61,    54,    30,    21,    15,     6,     3,    52,   210,   179,
   177,    33,   178,     8,   180,   179,   177,    64,   178,   273,
   180,    24,   123,    44,   233,   231,    93,   184,   182,   185,
    45,    78,    55,   184,   182,   185,    46,    34,    16,   255,
   135,    53,    41,    20,    14,     9,    39,     5,   162,    98,
   283,   196,   168,   138,   131,   117,   114,    11,   154,   149,
   155,   181,   226,   176,   200,    90,   181,   181,   141,   166,
    88,   129,    99,   110,   285,   179,   148,    22,   284,   257,
   180,    77,   280,    79,    73,   276,   275,   274,   272,   262,
   271,   270,   179,   177,   269,   178,   247,   180,   222,   245,
   212,    35,   179,   177,   241,   178,   240,   180,   230,   279,
   195,   210,   179,   177,   165,   178,   160,   180,   259,   278,
   140,   256,   139,   137,   147,   145,   136,   181,   157,    83,
   120,   261,   119,   118,   147,   145,   127,   125,   115,   214,
   221,   158,   211,   157,   181,   111,   179,   177,    96,   178,
    94,   180,    70,    81,   181,    67,    59,   109,   102,   103,
   104,    49,    25,   159,   181,   106,   144,    17,   267,   153,
    42,   244,   243,   151,   170,   171,   144,   121,   124,   105,
   242,   183,   186,   187,   192,   194,    74,   128,   193,    72,
    75,    31,     4,   133,     1,    10,     2,   188,   181,    71,
   163,    32,    85,    86,    87,    37,     7,   152,   201,   173,
   172,   174,   167,    26,    56,   190,    65,   197,    27,   189,
    18,    43,    12,    91,    47,   143,   183,   186,   187,    38,
   207,   208,   183,   186,   187,   215,   216,   217,   218,   219,
   220,    23,   199,   107,    51,     0,     0,     0,   224,     0,
   225,     0,   213,     0,   227,   227,     0,   229,     0,     0,
     0,   161,    97,   236,   239,   223,   234,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,   281,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,   265,   266,     0,     0,     0,     0,     0,     0,   254,
     0,     0,     0,     0,     0,   258,     0,   263 };
int yypact[]={

  -178, -1000,  -126,  -180,  -169,  -130,  -110,  -201,  -131,  -181,
  -138,   213,  -215,  -132,  -182,  -158,   193,  -110,  -221,  -227,
  -183,  -171,  -139,  -158,  -267, -1000, -1000,  -127,  -229,  -133,
  -149,  -140,  -171,  -212,   192, -1000,  -253, -1000,  -127,  -176,
  -134,  -184,  -144,  -149,  -269,  -270,   187, -1000,  -271, -1000,
  -185, -1000,  -272,  -186,  -223,   186, -1000,  -187,  -236, -1000,
   183,     0,  -188,  -223,  -145,  -223,  -231, -1000,   240,  -219,
 -1000,  -263,  -263,  -263,  -263,  -274,  -276,  -150,   181, -1000,
  -273,   179,    94,  -217,  -189,  -217,  -217,  -217,   228,   211,
  -198,  -276,   102,   176, -1000,  -190, -1000,  -242,  -112,   169,
  -191,  -113,   164,   163,   161, -1000,  -274,  -250,  -157, -1000,
   -31, -1000,  -265,  -114,  -243, -1000,    70, -1000, -1000, -1000,
 -1000, -1000,  -192,  -135,   157,   154,  -115,   153,   151,   -43,
   -40,   147,    93, -1000,  -276,  -193, -1000, -1000,   145, -1000,
 -1000,    66, -1000,   -33, -1000, -1000,  -116, -1000, -1000,  -291,
   -40,   -40, -1000,  -194,    63,    12,   -25, -1000,   238,   235,
 -1000,   141,  -117,  -199,  -158, -1000, -1000, -1000, -1000,  -211,
   -40,   -40,   -37,    57, -1000,   -67,   -25,   -25,   -25,   -25,
   -25,   -25, -1000, -1000, -1000, -1000, -1000, -1000,   -69,   -25,
    62,   -25, -1000,  -254,  -254, -1000,   139,  -151,  -222, -1000,
  -152,  -211,  -262,  -255,   137,   135,   230,  -295,  -294, -1000,
 -1000, -1000, -1000, -1000, -1000,   194,   123,   123,    62,    62,
    62, -1000, -1000, -1000,   194,   160,   221, -1000, -1000,   220,
 -1000,   130,  -233,   127, -1000,  -232,  -245,  -210,  -247,  -248,
 -1000, -1000,    70, -1000, -1000, -1000,  -136, -1000,   -88,  -278,
   -78,  -280,   -25,   -25,   217,  -195,   125,   122,   121,   119,
  -161,   118,   117,   116,  -251,   150,   140,   113,  -223, -1000,
 -1000, -1000, -1000,  -281, -1000, -1000, -1000,  -118, -1000, -1000,
 -1000, -1000,   109,   105, -1000, -1000 };
int yypgo[]={

     0,   333,   167,   331,   174,   160,   295,   319,   162,   166,
   158,   315,   155,   313,   312,   260,   311,   310,   308,   307,
   107,   306,   154,   298,   150,   152,   148,   149,   297,   153,
   296,   281,   291,   286,   285,   219,   284,   282 };
int yyr1[]={

     0,    36,    33,    33,    34,    34,    37,     2,     2,     3,
     3,     3,     3,    35,    35,     8,     8,     4,     5,     5,
    30,    31,    31,    32,    14,    14,    15,    15,    16,    16,
    16,    16,    17,    17,    18,    18,    20,    20,    21,    27,
    27,    27,    27,    27,    28,    28,    28,    28,    28,    28,
    28,    28,    28,    29,    29,    29,    29,    29,    29,    22,
    22,    23,    23,    23,    23,    23,    23,    23,    23,    23,
    23,    23,    23,    23,    23,    24,    24,    25,    25,    26,
    26,    26,    26,    26,    26,    26,    26,    26,     6,     6,
     7,     1,     1,    12,    12,    13,    13,    13,    13,    13,
     9,     9,    10,    10,    11,    11,    11,    11,    11,    19,
    19 };
int yyr2[]={

     0,    15,    11,     1,     3,     7,    13,     5,     3,    17,
    17,    17,    17,     7,     1,     7,     1,     7,     3,     7,
    13,     3,     5,     9,    13,     1,     5,     3,    11,    17,
    19,    21,    15,     1,    15,     1,     3,     5,    21,     7,
     7,     7,     5,     3,     7,     7,     7,     7,     7,     7,
     7,     7,     7,     3,     3,     3,     3,     3,     3,     3,
     5,    11,    11,    11,    11,    11,    11,    11,    11,    11,
    15,     5,     5,    11,    15,     9,     9,     3,     3,     7,
     7,     7,     7,     7,     7,     5,     3,     3,     3,     5,
    25,    11,     1,     3,     5,     9,     9,    11,     9,     9,
     7,     5,     3,     5,     3,     3,     5,     3,     3,    13,
     1 };
int yychk[]={

 -1000,   -36,   -33,   274,   -37,   263,   275,   -30,   272,   265,
   -34,   257,   -14,   278,   265,   275,   266,    44,   -17,   283,
   265,   275,    -2,    -3,   269,    59,   -34,   -18,   286,   288,
   275,   -31,   -32,   272,   266,    -2,   299,    -6,    -7,   263,
   288,   265,   -15,   -16,   262,   269,   266,   -31,   281,    59,
   295,    -6,   273,   265,   275,   266,   -15,   299,   299,    59,
   299,   275,   299,   275,   -20,   -21,   286,    59,   275,   290,
    59,   289,   279,    -4,   276,    40,   275,   -20,   266,   -20,
   288,    -4,   285,   -35,   297,   -35,   -35,   -35,    -5,   299,
   -12,   -13,   299,   266,    59,   299,    59,   258,    45,    -8,
   284,   275,    -8,    -8,    -8,    41,    44,    -1,   277,   -12,
    61,    59,   275,   294,   258,    59,   275,   258,    59,    59,
    59,    -5,   291,   269,   299,   258,    45,   257,    -9,    91,
   298,   258,   294,    -9,   275,   265,    59,    59,   258,    59,
    59,   -10,    93,   -11,   299,   258,    45,   257,    -9,   -27,
    40,   303,   -28,   299,   -26,   -24,    45,   258,   271,   293,
    59,   258,    45,   -12,   275,    59,    93,   -10,   258,   292,
   301,   302,   -27,   -26,   -27,   275,   -29,    43,    45,    42,
    47,    94,    61,   259,    60,    62,   260,   261,   275,   -29,
   -26,    40,   299,    40,    40,    59,   258,   -19,   277,    -2,
   -22,   -23,   264,   267,   286,   287,   282,   -27,   -27,    41,
    41,   299,   257,    -9,   296,   -26,   -26,   -26,   -26,   -26,
   -26,   299,   257,    -9,   -26,   -26,   -25,   -24,   299,   -25,
    59,   266,   286,   266,   -22,   299,   -24,   297,   299,   -24,
    59,    59,    40,    41,    41,    59,   288,    59,   294,   268,
   294,   280,   294,   294,    -9,   265,   299,   257,    -9,   296,
   299,   299,   257,    -9,   299,   -26,   -26,    41,   275,    59,
    59,    59,    59,   270,    59,    59,    59,   294,    59,    59,
    59,   -20,   299,   258,    59,    59 };
int yydef[]={

     3,    -2,     0,     0,     0,     0,     0,    25,     0,     0,
     0,     4,    33,     0,     0,     0,     0,     0,    35,     0,
     0,     0,     0,     8,     0,     2,     5,     0,     0,     0,
     0,     0,    21,     0,     0,     7,     0,     1,    88,     0,
     0,     0,     0,    27,     0,     0,     0,    22,     0,     6,
     0,    89,     0,     0,     0,     0,    26,     0,     0,    20,
     0,     0,     0,     0,     0,    36,     0,    24,     0,     0,
    23,    14,    14,    14,    14,     0,     0,     0,     0,    37,
     0,     0,     0,    16,     0,    16,    16,    16,     0,    18,
    92,    93,     0,     0,    32,     0,    28,     0,     0,     0,
     0,     0,     0,     0,     0,    17,     0,     0,     0,    94,
     0,    34,     0,     0,     0,     9,     0,    13,    10,    11,
    12,    19,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,    15,     0,     0,    95,    96,     0,    98,
    99,     0,   101,   102,   104,   105,     0,   107,   108,     0,
     0,     0,    43,    87,     0,     0,     0,    86,     0,     0,
    29,     0,     0,   110,     0,    97,   100,   103,   106,     0,
     0,     0,     0,     0,    42,     0,     0,     0,     0,     0,
     0,     0,    53,    54,    55,    56,    57,    58,     0,     0,
    85,     0,    87,     0,     0,    30,     0,     0,     0,    91,
     0,    59,     0,     0,     0,     0,     0,    39,    40,    41,
    84,    44,    46,    47,    52,    45,    79,    80,    81,    82,
    83,    48,    50,    51,    49,     0,     0,    77,    78,     0,
    31,     0,     0,     0,    60,     0,     0,     0,     0,     0,
    71,    72,     0,    75,    76,    90,     0,    38,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,    61,
    63,    64,    69,     0,    65,    67,    68,     0,    62,    66,
    73,   109,     0,     0,    70,    74 };
typedef struct { char *t_name; int t_val; } yytoktype;
#ifndef YYDEBUG
#	define YYDEBUG	0	/* don't allow debugging */
#endif

#if YYDEBUG

yytoktype yytoks[] =
{
	"STRING",	257,
	"REALNUM",	258,
	"NOTEQUAL",	259,
	"LESSOREQUAL",	260,
	"MOREOREQUAL",	261,
	"=",	61,
	"<",	60,
	">",	62,
	"ABSTRACTION",	262,
	"CASE",	263,
	"CHANGE",	264,
	"DEFINITION",	265,
	"END",	266,
	"EVALUATE",	267,
	"EXTRACT",	268,
	"FIELD",	269,
	"FROM",	270,
	"HD",	271,
	"INDEX",	272,
	"INSTANCE",	273,
	"INTRODUCTION",	274,
	"IS",	275,
	"LIST",	276,
	"LOCAL",	277,
	"MODIFICATION",	278,
	"NUMBER",	279,
	"OF",	280,
	"ON",	281,
	"PR",	282,
	"PREPROCESS",	283,
	"PROMPT",	284,
	"RANGE",	285,
	"REPAIR",	286,
	"RESELECT",	287,
	"RULE",	288,
	"KSTRING",	289,
	"SIMILAR",	290,
	"SOLUTION",	291,
	"THEN",	292,
	"TL",	293,
	"TO",	294,
	"TYPE",	295,
	"UNDEFINED",	296,
	"WEIGHT",	297,
	"WHEN",	298,
	"IDENTIFIER",	299,
	"+",	43,
	"-",	45,
	"*",	42,
	"/",	47,
	"UNARYMINUS",	300,
	"^",	94,
	"OR",	301,
	"AND",	302,
	"NOT",	303,
	"-unknown-",	-1	/* ends search */
};

char * yyreds[] =
{
	"-no such reduction-",
	"caseseq : introblock defns indexblock modblock preprocessblock repblock cases",
	"introblock : INTRODUCTION IS introseq END ';'",
	"introblock : /* empty */",
	"introseq : STRING",
	"introseq : STRING ',' introseq",
	"defns : CASE DEFINITION IS fielddefs END ';'",
	"fielddefs : fielddef fielddefs",
	"fielddefs : fielddef",
	"fielddef : FIELD IDENTIFIER TYPE IS KSTRING weightdef promptdef ';'",
	"fielddef : FIELD IDENTIFIER TYPE IS NUMBER weightdef promptdef ';'",
	"fielddef : FIELD IDENTIFIER TYPE IS enumeration weightdef promptdef ';'",
	"fielddef : FIELD IDENTIFIER TYPE IS LIST weightdef promptdef ';'",
	"weightdef : WEIGHT IS REALNUM",
	"weightdef : /* empty */",
	"promptdef : PROMPT IS listbox",
	"promptdef : /* empty */",
	"enumeration : '(' identlist ')'",
	"identlist : IDENTIFIER",
	"identlist : IDENTIFIER ',' identlist",
	"indexblock : INDEX DEFINITION IS indexlist END ';'",
	"indexlist : index",
	"indexlist : index indexlist",
	"index : INDEX ON IDENTIFIER ';'",
	"modblock : MODIFICATION DEFINITION IS modrules END ';'",
	"modblock : /* empty */",
	"modrules : modrule modrules",
	"modrules : modrule",
	"modrule : ABSTRACTION IDENTIFIER IS enumeration ';'",
	"modrule : FIELD IDENTIFIER SIMILAR RANGE REALNUM TO REALNUM ';'",
	"modrule : FIELD IDENTIFIER SIMILAR RANGE '-' REALNUM TO REALNUM ';'",
	"modrule : FIELD IDENTIFIER SIMILAR RANGE '-' REALNUM TO '-' REALNUM ';'",
	"preprocessblock : PREPROCESS RULE DEFINITION IS repseq END ';'",
	"preprocessblock : /* empty */",
	"repblock : REPAIR RULE DEFINITION IS repseq END ';'",
	"repblock : /* empty */",
	"repseq : reprule",
	"repseq : reprule repseq",
	"reprule : REPAIR RULE IDENTIFIER IS WHEN combtree THEN replist END ';'",
	"combtree : combtree OR combtree",
	"combtree : combtree AND combtree",
	"combtree : '(' combtree ')'",
	"combtree : NOT combtree",
	"combtree : combtreeitem",
	"combtreeitem : IDENTIFIER IS IDENTIFIER",
	"combtreeitem : expression compop expression",
	"combtreeitem : IDENTIFIER IS STRING",
	"combtreeitem : IDENTIFIER IS listbox",
	"combtreeitem : listop IS IDENTIFIER",
	"combtreeitem : listop compop expression",
	"combtreeitem : listop IS STRING",
	"combtreeitem : listop IS listbox",
	"combtreeitem : IDENTIFIER IS UNDEFINED",
	"compop : '='",
	"compop : NOTEQUAL",
	"compop : '<'",
	"compop : '>'",
	"compop : LESSOREQUAL",
	"compop : MOREOREQUAL",
	"replist : replistitem",
	"replist : replistitem replist",
	"replistitem : CHANGE IDENTIFIER TO IDENTIFIER ';'",
	"replistitem : EVALUATE IDENTIFIER TO expression ';'",
	"replistitem : CHANGE IDENTIFIER TO STRING ';'",
	"replistitem : CHANGE IDENTIFIER TO listbox ';'",
	"replistitem : CHANGE listop TO IDENTIFIER ';'",
	"replistitem : EVALUATE listop TO expression ';'",
	"replistitem : CHANGE listop TO STRING ';'",
	"replistitem : CHANGE listop TO listbox ';'",
	"replistitem : CHANGE IDENTIFIER TO UNDEFINED ';'",
	"replistitem : CHANGE IDENTIFIER EXTRACT IDENTIFIER FROM IDENTIFIER ';'",
	"replistitem : REPAIR ';'",
	"replistitem : RESELECT ';'",
	"replistitem : PR '(' listbox ')' ';'",
	"replistitem : CHANGE WEIGHT OF IDENTIFIER TO REALNUM ';'",
	"listop : HD '(' listop2 ')'",
	"listop : TL '(' listop2 ')'",
	"listop2 : listop",
	"listop2 : IDENTIFIER",
	"expression : expression '+' expression",
	"expression : expression '-' expression",
	"expression : expression '*' expression",
	"expression : expression '/' expression",
	"expression : expression '^' expression",
	"expression : '(' expression ')'",
	"expression : '-' expression",
	"expression : REALNUM",
	"expression : IDENTIFIER",
	"cases : case",
	"cases : case cases",
	"case : CASE INSTANCE IDENTIFIER IS fieldlist local_fielddefs SOLUTION IS fieldlist local_repairs END ';'",
	"local_fielddefs : LOCAL FIELD DEFINITION IS fielddefs",
	"local_fielddefs : /* empty */",
	"fieldlist : field",
	"fieldlist : field fieldlist",
	"field : IDENTIFIER '=' IDENTIFIER ';'",
	"field : IDENTIFIER '=' REALNUM ';'",
	"field : IDENTIFIER '=' '-' REALNUM ';'",
	"field : IDENTIFIER '=' STRING ';'",
	"field : IDENTIFIER '=' listbox ';'",
	"listbox : '[' list ']'",
	"listbox : '[' ']'",
	"list : listitem",
	"list : listitem list",
	"listitem : IDENTIFIER",
	"listitem : REALNUM",
	"listitem : '-' REALNUM",
	"listitem : STRING",
	"listitem : listbox",
	"local_repairs : LOCAL REPAIR RULE DEFINITION IS repseq",
	"local_repairs : /* empty */",
};
#endif /* YYDEBUG */
#line 1 "/usr/lib/yaccpar"
/*	@(#)yaccpar 1.10 89/04/04 SMI; from S5R3 1.10	*/

/*
** Skeleton parser driver for yacc output
*/

/*
** yacc user known macros and defines
*/
#define YYERROR		goto yyerrlab
#define YYACCEPT	{ free(yys); free(yyv); return(0); }
#define YYABORT		{ free(yys); free(yyv); return(1); }
#define YYBACKUP( newtoken, newvalue )\
{\
	if ( yychar >= 0 || ( yyr2[ yytmp ] >> 1 ) != 1 )\
	{\
		yyerror( "syntax error - cannot backup" );\
		goto yyerrlab;\
	}\
	yychar = newtoken;\
	yystate = *yyps;\
	yylval = newvalue;\
	goto yynewstate;\
}
#define YYRECOVERING()	(!!yyerrflag)
#ifndef YYDEBUG
#	define YYDEBUG	1	/* make debugging available */
#endif

/*
** user known globals
*/
int yydebug;			/* set to 1 to get debugging */

/*
** driver internal defines
*/
#define YYFLAG		(-1000)

/*
** static variables used by the parser
*/
static YYSTYPE *yyv;			/* value stack */
static int *yys;			/* state stack */

static YYSTYPE *yypv;			/* top of value stack */
static int *yyps;			/* top of state stack */

static int yystate;			/* current state */
static int yytmp;			/* extra var (lasts between blocks) */

int yynerrs;			/* number of errors */

int yyerrflag;			/* error recovery flag */
int yychar;			/* current input token number */


/*
** yyparse - return 0 if worked, 1 if syntax error not recovered from
*/
int
yyparse()
{
	register YYSTYPE *yypvt;	/* top of value stack for $vars */
	unsigned yymaxdepth = YYMAXDEPTH;

	/*
	** Initialize externals - yyparse may be called more than once
	*/
	yyv = (YYSTYPE*)malloc(yymaxdepth*sizeof(YYSTYPE));
	yys = (int*)malloc(yymaxdepth*sizeof(int));
	if (!yyv || !yys)
	{
		yyerror( "out of memory" );
		return(1);
	}
	yypv = &yyv[-1];
	yyps = &yys[-1];
	yystate = 0;
	yytmp = 0;
	yynerrs = 0;
	yyerrflag = 0;
	yychar = -1;

	goto yystack;
	{
		register YYSTYPE *yy_pv;	/* top of value stack */
		register int *yy_ps;		/* top of state stack */
		register int yy_state;		/* current state */
		register int  yy_n;		/* internal state number info */

		/*
		** get globals into registers.
		** branch to here only if YYBACKUP was called.
		*/
	yynewstate:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;
		goto yy_newstate;

		/*
		** get globals into registers.
		** either we just started, or we just finished a reduction
		*/
	yystack:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;

		/*
		** top of for (;;) loop while no reductions done
		*/
	yy_stack:
		/*
		** put a state and value onto the stacks
		*/
#if YYDEBUG
		/*
		** if debugging, look up token value in list of value vs.
		** name pairs.  0 and negative (-1) are special values.
		** Note: linear search is used since time is not a real
		** consideration while debugging.
		*/
		if ( yydebug )
		{
			register int yy_i;

			(void)printf( "State %d, token ", yy_state );
			if ( yychar == 0 )
				(void)printf( "end-of-file\n" );
			else if ( yychar < 0 )
				(void)printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				(void)printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ++yy_ps >= &yys[ yymaxdepth ] )	/* room on stack? */
		{
			/*
			** reallocate and recover.  Note that pointers
			** have to be reset, or bad things will happen
			*/
			int yyps_index = (yy_ps - yys);
			int yypv_index = (yy_pv - yyv);
			int yypvt_index = (yypvt - yyv);
			yymaxdepth += YYMAXDEPTH;
			yyv = (YYSTYPE*)realloc((char*)yyv,
				yymaxdepth * sizeof(YYSTYPE));
			yys = (int*)realloc((char*)yys,
				yymaxdepth * sizeof(int));
			if (!yyv || !yys)
			{
				yyerror( "yacc stack overflow" );
				return(1);
			}
			yy_ps = yys + yyps_index;
			yy_pv = yyv + yypv_index;
			yypvt = yyv + yypvt_index;
		}
		*yy_ps = yy_state;
		*++yy_pv = yyval;

		/*
		** we have a new state - find out what to do
		*/
	yy_newstate:
		if ( ( yy_n = yypact[ yy_state ] ) <= YYFLAG )
			goto yydefault;		/* simple state */
#if YYDEBUG
		/*
		** if debugging, need to mark whether new token grabbed
		*/
		yytmp = yychar < 0;
#endif
		if ( ( yychar < 0 ) && ( ( yychar = yylex() ) < 0 ) )
			yychar = 0;		/* reached EOF */
#if YYDEBUG
		if ( yydebug && yytmp )
		{
			register int yy_i;

			(void)printf( "Received token " );
			if ( yychar == 0 )
				(void)printf( "end-of-file\n" );
			else if ( yychar < 0 )
				(void)printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				(void)printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ( ( yy_n += yychar ) < 0 ) || ( yy_n >= YYLAST ) )
			goto yydefault;
		if ( yychk[ yy_n = yyact[ yy_n ] ] == yychar )	/*valid shift*/
		{
			yychar = -1;
			yyval = yylval;
			yy_state = yy_n;
			if ( yyerrflag > 0 )
				yyerrflag--;
			goto yy_stack;
		}

	yydefault:
		if ( ( yy_n = yydef[ yy_state ] ) == -2 )
		{
#if YYDEBUG
			yytmp = yychar < 0;
#endif
			if ( ( yychar < 0 ) && ( ( yychar = yylex() ) < 0 ) )
				yychar = 0;		/* reached EOF */
#if YYDEBUG
			if ( yydebug && yytmp )
			{
				register int yy_i;

				(void)printf( "Received token " );
				if ( yychar == 0 )
					(void)printf( "end-of-file\n" );
				else if ( yychar < 0 )
					(void)printf( "-none-\n" );
				else
				{
					for ( yy_i = 0;
						yytoks[yy_i].t_val >= 0;
						yy_i++ )
					{
						if ( yytoks[yy_i].t_val
							== yychar )
						{
							break;
						}
					}
					(void)printf( "%s\n", yytoks[yy_i].t_name );
				}
			}
#endif /* YYDEBUG */
			/*
			** look through exception table
			*/
			{
				register int *yyxi = yyexca;

				while ( ( *yyxi != -1 ) ||
					( yyxi[1] != yy_state ) )
				{
					yyxi += 2;
				}
				while ( ( *(yyxi += 2) >= 0 ) &&
					( *yyxi != yychar ) )
					;
				if ( ( yy_n = yyxi[1] ) < 0 )
					YYACCEPT;
			}
		}

		/*
		** check for syntax error
		*/
		if ( yy_n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( yyerrflag )
			{
			case 0:		/* new error */
				yyerror( "syntax error" );
				goto skip_init;
			yyerrlab:
				/*
				** get globals into registers.
				** we have a user generated syntax type error
				*/
				yy_pv = yypv;
				yy_ps = yyps;
				yy_state = yystate;
				yynerrs++;
			skip_init:
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				yyerrflag = 3;
				/*
				** find state where "error" is a legal
				** shift action
				*/
				while ( yy_ps >= yys )
				{
					yy_n = yypact[ *yy_ps ] + YYERRCODE;
					if ( yy_n >= 0 && yy_n < YYLAST &&
						yychk[yyact[yy_n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						yy_state = yyact[ yy_n ];
						goto yy_stack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
#if YYDEBUG
#	define _POP_ "Error recovery pops state %d, uncovers state %d\n"
					if ( yydebug )
						(void)printf( _POP_, *yy_ps,
							yy_ps[-1] );
#	undef _POP_
#endif
					yy_ps--;
					yy_pv--;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
				YYABORT;
			case 3:		/* no shift yet; eat a token */
#if YYDEBUG
				/*
				** if debugging, look up token in list of
				** pairs.  0 and negative shouldn't occur,
				** but since timing doesn't matter when
				** debugging, it doesn't hurt to leave the
				** tests here.
				*/
				if ( yydebug )
				{
					register int yy_i;

					(void)printf( "Error recovery discards " );
					if ( yychar == 0 )
						(void)printf( "token end-of-file\n" );
					else if ( yychar < 0 )
						(void)printf( "token -none-\n" );
					else
					{
						for ( yy_i = 0;
							yytoks[yy_i].t_val >= 0;
							yy_i++ )
						{
							if ( yytoks[yy_i].t_val
								== yychar )
							{
								break;
							}
						}
						(void)printf( "token %s\n",
							yytoks[yy_i].t_name );
					}
				}
#endif /* YYDEBUG */
				if ( yychar == 0 )	/* reached EOF. quit */
					YYABORT;
				yychar = -1;
				goto yy_newstate;
			}
		}/* end if ( yy_n == 0 ) */
		/*
		** reduction by production yy_n
		** put stack tops, etc. so things right after switch
		*/
#if YYDEBUG
		/*
		** if debugging, print the string that is the user's
		** specification of the reduction which is just about
		** to be done.
		*/
		if ( yydebug )
			(void)printf( "Reduce by (%d) \"%s\"\n",
				yy_n, yyreds[ yy_n ] );
#endif
		yytmp = yy_n;			/* value to switch over */
		yypvt = yy_pv;			/* $vars top of value stack */
		/*
		** Look in goto table for next state
		** Sorry about using yy_state here as temporary
		** register variable, but why not, if it works...
		** If yyr2[ yy_n ] doesn't have the low order bit
		** set, then there is no action to be done for
		** this reduction.  So, no saving & unsaving of
		** registers done.  The only difference between the
		** code just after the if and the body of the if is
		** the goto yy_stack in the body.  This way the test
		** can be made before the choice of what to do is needed.
		*/
		{
			/* length of production doubled with extra bit */
			register int yy_len = yyr2[ yy_n ];

			if ( !( yy_len & 01 ) )
			{
				yy_len >>= 1;
				yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
				yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
					*( yy_ps -= yy_len ) + 1;
				if ( yy_state >= YYLAST ||
					yychk[ yy_state =
					yyact[ yy_state ] ] != -yy_n )
				{
					yy_state = yyact[ yypgo[ yy_n ] ];
				}
				goto yy_stack;
			}
			yy_len >>= 1;
			yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
			yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
				*( yy_ps -= yy_len ) + 1;
			if ( yy_state >= YYLAST ||
				yychk[ yy_state = yyact[ yy_state ] ] != -yy_n )
			{
				yy_state = yyact[ yypgo[ yy_n ] ];
			}
		}
					/* save until reenter driver code */
		yystate = yy_state;
		yyps = yy_ps;
		yypv = yy_pv;
	}
	/*
	** code supplied by user is placed in this switch
	*/
	switch( yytmp )
	{
		
case 1:
# line 592 "cases49"
{
				 introroot = yypvt[-6].introptr;
				 indexnameroot = yypvt[-4].indexnameptr;
				 modroot = yypvt[-3].modptr;
				 preprocessroot = yypvt[-2].repptr;
				 reproot = yypvt[-1].repptr;
				 caseroot = yypvt[-0].caseptr;
				} break;
case 2:
# line 603 "cases49"
{
				 yyval.introptr = yypvt[-2].introptr;
				} break;
case 3:
# line 607 "cases49"
{
				 yyval.introptr = NULL;
				} break;
case 4:
# line 613 "cases49"
{
				 yyval.introptr = createintronode
					(
					 yypvt[-0].strval
					);
				} break;
case 5:
# line 620 "cases49"
{
				 yyval.introptr = createintronode
					(
					 yypvt[-2].strval
					);
				 yyval.introptr->introlink = yypvt[-0].introptr;
				} break;
case 6:
# line 630 "cases49"
{
				 probdefnroot = yypvt[-2].typeptr;
				} break;
case 7:
# line 636 "cases49"
{
				 yyval.typeptr = yypvt[-1].typeptr;
				 yyval.typeptr->typelink = yypvt[-0].typeptr;
				} break;
case 8:
# line 641 "cases49"
{
				 yyval.typeptr = yypvt[-0].typeptr;
				} break;
case 9:
# line 647 "cases49"
{
				 yyval.typeptr = createtypenode
					(
					 yypvt[-6].strval,
					 STRTYPE,
					 yypvt[-2].realval,
					 yypvt[-1].listptr
					);
				 free(yypvt[-6].strval);
				} break;
case 10:
# line 658 "cases49"
{
				 yyval.typeptr = createtypenode
					(
					 yypvt[-6].strval,
					 NUMTYPE,
					 yypvt[-2].realval,
					 yypvt[-1].listptr
					);
				 free(yypvt[-6].strval);
				} break;
case 11:
# line 669 "cases49"
{
				 yyval.typeptr = createtypenode
					(
					 yypvt[-6].strval,
					 ENUMTYPE,
					 yypvt[-2].realval,
					 yypvt[-1].listptr
					);
				 yyval.typeptr->enumhead = yypvt[-3].enumptr;
				 free(yypvt[-6].strval);
				} break;
case 12:
# line 681 "cases49"
{
				 yyval.typeptr = createtypenode
					(
					 yypvt[-6].strval,
					 LISTTYPE,
					 yypvt[-2].realval,
					 yypvt[-1].listptr
					);
				 free(yypvt[-6].strval);
				} break;
case 13:
# line 694 "cases49"
{
				 yyval.realval = yypvt[-0].realval;
				} break;
case 14:
# line 698 "cases49"
{
				 yyval.realval = 1.0;
				} break;
case 15:
# line 704 "cases49"
{
				 yyval.listptr = yypvt[-0].listptr;
				} break;
case 16:
# line 708 "cases49"
{
				 yyval.listptr = NULL;
				} break;
case 17:
# line 714 "cases49"
{
				 yyval.enumptr = yypvt[-1].enumptr;
				} break;
case 18:
# line 720 "cases49"
{
				 yyval.enumptr = createenumnode
					(
					 yypvt[-0].strval
					);
				 free(yypvt[-0].strval);
				} break;
case 19:
# line 728 "cases49"
{
				 yyval.enumptr = createenumnode
                                        (
                                         yypvt[-2].strval
                                        );
				 yyval.enumptr->enumlink = yypvt[-0].enumptr;
				 free(yypvt[-2].strval);
				} break;
case 20:
# line 739 "cases49"
{
				 yyval.indexnameptr = yypvt[-2].indexnameptr;
				} break;
case 21:
# line 745 "cases49"
{
				 yyval.indexnameptr = yypvt[-0].indexnameptr;
				} break;
case 22:
# line 749 "cases49"
{
				 yyval.indexnameptr = yypvt[-1].indexnameptr;
				 yyval.indexnameptr->indexnamelink = yypvt[-0].indexnameptr;
				} break;
case 23:
# line 756 "cases49"
{
				 yyval.indexnameptr = createindexnamenode
					(
					 yypvt[-1].strval
					);
				 free(yypvt[-1].strval);
				} break;
case 24:
# line 766 "cases49"
{
				 yyval.modptr = yypvt[-2].modptr;
				} break;
case 25:
# line 770 "cases49"
{
				 yyval.modptr = NULL;
				} break;
case 26:
# line 776 "cases49"
{
				 yyval.modptr = yypvt[-1].modptr;
				 yyval.modptr->modlink = yypvt[-0].modptr;
				} break;
case 27:
# line 781 "cases49"
{
				 yyval.modptr = yypvt[-0].modptr;
				} break;
case 28:
# line 787 "cases49"
{
				 yyval.modptr = createmodenumnode
					(
					 yypvt[-3].strval,
					 yypvt[-1].enumptr
					);
				 free(yypvt[-3].strval);
				} break;
case 29:
# line 796 "cases49"
{
				 if (yypvt[-1].realval < yypvt[-3].realval)
				   {
				    yyerror("Invalid number range");
				   }

				 yyval.modptr = createmodrangenode
					(
					 yypvt[-6].strval,
					 yypvt[-3].realval,
					 yypvt[-1].realval
					);
				 free(yypvt[-6].strval);
				} break;
case 30:
# line 811 "cases49"
{
				 yyval.modptr = createmodrangenode
					(
					 yypvt[-7].strval,
					 -yypvt[-3].realval,
					 yypvt[-1].realval
					);
				 free(yypvt[-7].strval);
				} break;
case 31:
# line 821 "cases49"
{
				 if (-yypvt[-1].realval < -yypvt[-4].realval)
				   {
				    yyerror("Invalid number range");
				   }

				 yyval.modptr = createmodrangenode
					(
					 yypvt[-8].strval,
					 -yypvt[-4].realval,
					 -yypvt[-1].realval
					);
				 free(yypvt[-8].strval);
				} break;
case 32:
# line 838 "cases49"
{
				 yyval.repptr = yypvt[-2].repptr;
				} break;
case 33:
# line 842 "cases49"
{
				 yyval.repptr = NULL;
				} break;
case 34:
# line 848 "cases49"
{
				 yyval.repptr = yypvt[-2].repptr;
				} break;
case 35:
# line 852 "cases49"
{
				 yyval.repptr = NULL;
				} break;
case 36:
# line 858 "cases49"
{
				 yyval.repptr = yypvt[-0].repptr;
				} break;
case 37:
# line 862 "cases49"
{
				 yyval.repptr = yypvt[-1].repptr;
				 yyval.repptr->replink = yypvt[-0].repptr;
				} break;
case 38:
# line 869 "cases49"
{
				 yyval.repptr = createrepnode
					(
					 yypvt[-7].strval,
					 yypvt[-4].combtreeptr,
					 yypvt[-2].replistptr
					);
				 free(yypvt[-7].strval);
				} break;
case 39:
# line 881 "cases49"
{
				 yyval.combtreeptr = createcombtreenode
					(
					 OR_OP,
					 yypvt[-2].combtreeptr,
					 yypvt[-0].combtreeptr
					);
				} break;
case 40:
# line 890 "cases49"
{
				 yyval.combtreeptr = createcombtreenode
					(
					 AND_OP,
					 yypvt[-2].combtreeptr,
					 yypvt[-0].combtreeptr
					);
				} break;
case 41:
# line 899 "cases49"
{
				 yyval.combtreeptr = yypvt[-1].combtreeptr;
				} break;
case 42:
# line 903 "cases49"
{
				 yyval.combtreeptr = createcombtreenode
					(
					 NOT_OP,
					 yypvt[-0].combtreeptr,
					 NULL
					);
				} break;
case 43:
# line 912 "cases49"
{
				 yyval.combtreeptr = createcombtreenode
					(
					 ITEM,
					 yypvt[-0].combptr,
					 NULL
					);
				} break;
case 44:
# line 923 "cases49"
{
				 yyval.combptr = createcombenumnode
					(
					 yypvt[-2].strval,
					 yypvt[-0].strval
					);
				 free(yypvt[-2].strval);
				 free(yypvt[-0].strval);
				} break;
case 45:
# line 933 "cases49"
{
				 yyval.combptr = createcombcompnode
					(
					 yypvt[-2].arithptr,
					 yypvt[-1].comptype,
					 yypvt[-0].arithptr
					);
				} break;
case 46:
# line 942 "cases49"
{
				 yyval.combptr = createcombstringnode
					(
					 yypvt[-2].strval,
					 yypvt[-0].strval
					);
				 free(yypvt[-2].strval);
				} break;
case 47:
# line 951 "cases49"
{
				 yyval.combptr = createcomblistnode
					(
					 yypvt[-2].strval,
					 yypvt[-0].listptr
					);
				 free(yypvt[-2].strval);
				} break;
case 48:
# line 960 "cases49"
{
				 yyval.combptr = createcomblistopenumnode
					(
					 yypvt[-2].listopptr,
					 yypvt[-0].strval
					);
				 free(yypvt[-0].strval);
				} break;
case 49:
# line 969 "cases49"
{
				 yyval.combptr = createcomblistopexpnode
					(
					 yypvt[-2].listopptr,
					 yypvt[-1].comptype,
					 yypvt[-0].arithptr
					);
				} break;
case 50:
# line 978 "cases49"
{
				 yyval.combptr = createcomblistopstringnode
					(
					 yypvt[-2].listopptr,
					 yypvt[-0].strval
					);
				 free(yypvt[-0].strval);
				} break;
case 51:
# line 987 "cases49"
{
				 yyval.combptr = createcomblistoplistnode
					(
					 yypvt[-2].listopptr,
					 yypvt[-0].listptr
					);
				} break;
case 52:
# line 995 "cases49"
{
				 yyval.combptr = createcombundefnode
					(
					 yypvt[-2].strval
					);
				 free(yypvt[-2].strval);
				} break;
case 53:
# line 1006 "cases49"
{
				 yyval.comptype = yypvt[-0].comptype;
				} break;
case 54:
# line 1010 "cases49"
{
				 yyval.comptype = yypvt[-0].comptype;
				} break;
case 55:
# line 1014 "cases49"
{
				 yyval.comptype = yypvt[-0].comptype;
				} break;
case 56:
# line 1018 "cases49"
{
				 yyval.comptype = yypvt[-0].comptype;
				} break;
case 57:
# line 1022 "cases49"
{
				 yyval.comptype = yypvt[-0].comptype;
				} break;
case 58:
# line 1026 "cases49"
{
				 yyval.comptype = yypvt[-0].comptype;
				} break;
case 59:
# line 1033 "cases49"
{
				 yyval.replistptr = yypvt[-0].replistptr;
				} break;
case 60:
# line 1037 "cases49"
{
				 yyval.replistptr = yypvt[-1].replistptr;
				 yyval.replistptr->replistlink = yypvt[-0].replistptr;
				} break;
case 61:
# line 1044 "cases49"
{
				 yyval.replistptr = createreplistenumnode
					(
					 yypvt[-3].strval,
					 yypvt[-1].strval
					);
				 free(yypvt[-3].strval);
				} break;
case 62:
# line 1053 "cases49"
{
				 yyval.replistptr = createreplistexpnode
					(
					 yypvt[-3].strval,
					 yypvt[-1].arithptr
					);
				 free(yypvt[-3].strval);
				} break;
case 63:
# line 1062 "cases49"
{
				 yyval.replistptr = createrepliststringnode
					(
					 yypvt[-3].strval,
					 yypvt[-1].strval
					);
				 free(yypvt[-3].strval);
				} break;
case 64:
# line 1071 "cases49"
{
				 yyval.replistptr = createreplistlistnode
					(
					 yypvt[-3].strval,
					 yypvt[-1].listptr
					);
				 free(yypvt[-3].strval);
				} break;
case 65:
# line 1080 "cases49"
{
				 yyval.replistptr = createreplistlistopenumnode
					(
					 yypvt[-3].listopptr,
					 yypvt[-1].strval
					);
				} break;
case 66:
# line 1088 "cases49"
{
				 yyval.replistptr = createreplistlistopexpnode
					(
					 yypvt[-3].listopptr,
					 yypvt[-1].arithptr
					);
				} break;
case 67:
# line 1096 "cases49"
{
				 yyval.replistptr = createreplistlistopstringnode
					(
					 yypvt[-3].listopptr,
					 yypvt[-1].strval
					);
				} break;
case 68:
# line 1104 "cases49"
{
				 yyval.replistptr = createreplistlistoplistnode
					(
					 yypvt[-3].listopptr,
					 yypvt[-1].listptr
					);
				} break;
case 69:
# line 1112 "cases49"
{
				 yyval.replistptr = createreplistundefnode
					(
					 yypvt[-3].strval
					);
				 free(yypvt[-3].strval);
				} break;
case 70:
# line 1120 "cases49"
{
				 yyval.replistptr = createreplistextractnode
					(
					 yypvt[-5].strval,
					 yypvt[-3].strval,
					 yypvt[-1].strval
					);
				 free(yypvt[-5].strval);
				} break;
case 71:
# line 1130 "cases49"
{
				 yyval.replistptr = createreplistrepnode();
				} break;
case 72:
# line 1134 "cases49"
{
				 yyval.replistptr = createreplistresnode();
				} break;
case 73:
# line 1138 "cases49"
{
				 yyval.replistptr = createreplistprintnode
					(
					 yypvt[-2].listptr
					);
				} break;
case 74:
# line 1145 "cases49"
{
				 yyval.replistptr = createreplistweightnode
					(
					 yypvt[-3].strval,
					 yypvt[-1].realval
					);
				} break;
case 75:
# line 1155 "cases49"
{
				 yyval.listopptr = createlistopnode
					(
					 HEAD,
					 yypvt[-1].listopptr
					);
				} break;
case 76:
# line 1163 "cases49"
{
				 yyval.listopptr = createlistopnode
					(
					 TAIL,
					 yypvt[-1].listopptr
					);
				} break;
case 77:
# line 1173 "cases49"
{
				 yyval.listopptr = yypvt[-0].listopptr;
				} break;
case 78:
# line 1177 "cases49"
{
				 yyval.listopptr = createlistopfieldnode
					(
				 	 yypvt[-0].strval
					);
				} break;
case 79:
# line 1186 "cases49"
{
				 yyval.arithptr = createplusnode
					(
					 yypvt[-2].arithptr,
					 yypvt[-0].arithptr
					);
				} break;
case 80:
# line 1194 "cases49"
{
				 yyval.arithptr = createminusnode
					(
					 yypvt[-2].arithptr,
					 yypvt[-0].arithptr
					);
				} break;
case 81:
# line 1202 "cases49"
{
				 yyval.arithptr = createmultiplynode
					(
					 yypvt[-2].arithptr,
					 yypvt[-0].arithptr
					);
				} break;
case 82:
# line 1210 "cases49"
{
				 yyval.arithptr = createdividenode
					(
					 yypvt[-2].arithptr,
					 yypvt[-0].arithptr
					);
				} break;
case 83:
# line 1218 "cases49"
{
				 yyval.arithptr = createpowernode
					(
					 yypvt[-2].arithptr,
					 yypvt[-0].arithptr
					);
				} break;
case 84:
# line 1226 "cases49"
{
				 yyval.arithptr = yypvt[-1].arithptr;
				} break;
case 85:
# line 1230 "cases49"
{
				 yyval.arithptr = createunaryminusnode
					(
					 yypvt[-0].arithptr
					);
				} break;
case 86:
# line 1237 "cases49"
{
				 yyval.arithptr = createnumbernode
					(
					 yypvt[-0].realval
					);
				} break;
case 87:
# line 1244 "cases49"
{
				 yyval.arithptr = createfieldvaluenode
					(
					 yypvt[-0].strval
					);
				 free(yypvt[-0].strval);
				} break;
case 88:
# line 1254 "cases49"
{
				 yyval.caseptr = yypvt[-0].caseptr;
				} break;
case 89:
# line 1258 "cases49"
{
				 yyval.caseptr = yypvt[-1].caseptr;
				 yyval.caseptr->caselink = yypvt[-0].caseptr;
				} break;
case 90:
# line 1270 "cases49"
{
                                 yyval.caseptr = createcasenode
                                        (
                                         yypvt[-9].strval
                                        );
				 yyval.caseptr->fieldhead = yypvt[-7].fieldptr;
                                 yyval.caseptr->fielddefhead = yypvt[-6].typeptr;
                                 yyval.caseptr->solnhead = yypvt[-3].fieldptr;
                                 yyval.caseptr->rephead = yypvt[-2].repptr;
                                 free(yypvt[-9].strval);
                                } break;
case 91:
# line 1284 "cases49"
{
				 yyval.typeptr = yypvt[-0].typeptr;
				} break;
case 92:
# line 1288 "cases49"
{
				 yyval.typeptr = NULL;
				} break;
case 93:
# line 1294 "cases49"
{
				 yyval.fieldptr = yypvt[-0].fieldptr;
				} break;
case 94:
# line 1298 "cases49"
{
				 yyval.fieldptr = yypvt[-1].fieldptr;
				 yyval.fieldptr->fieldlink = yypvt[-0].fieldptr;
				} break;
case 95:
# line 1305 "cases49"
{
				 yyval.fieldptr = createenumfield
					(
					 yypvt[-3].strval,
					 yypvt[-1].strval
					);
				 free(yypvt[-3].strval);
				} break;
case 96:
# line 1314 "cases49"
{
				 yyval.fieldptr = createnumfield
					(
					 yypvt[-3].strval,
					 yypvt[-1].realval
					);
				 free(yypvt[-3].strval);
				} break;
case 97:
# line 1323 "cases49"
{
				 yyval.fieldptr = createnumfield
					(
					 yypvt[-4].strval,
					 -yypvt[-1].realval
					);
				 free(yypvt[-4].strval);
				} break;
case 98:
# line 1332 "cases49"
{
				 yyval.fieldptr = createstrfield
					(
					 yypvt[-3].strval,
					 yypvt[-1].strval
					);
				 free(yypvt[-3].strval);
				} break;
case 99:
# line 1341 "cases49"
{
				 yyval.fieldptr = createlistfield
					(
					 yypvt[-3].strval,
					 yypvt[-1].listptr
					);
				 free(yypvt[-3].strval);
				} break;
case 100:
# line 1352 "cases49"
{
				 yyval.listptr = yypvt[-1].listptr;
				} break;
case 101:
# line 1356 "cases49"
{
				 yyval.listptr = NULL;
				} break;
case 102:
# line 1362 "cases49"
{
				 yyval.listptr = yypvt[-0].listptr;
				 yyval.listptr->listlink = NULL;
				} break;
case 103:
# line 1367 "cases49"
{
				 yyval.listptr = yypvt[-1].listptr;
				 yyval.listptr->listlink = yypvt[-0].listptr;
				} break;
case 104:
# line 1374 "cases49"
{
				 yyval.listptr = createlistenum
					(
					 yypvt[-0].strval
					);
				} break;
case 105:
# line 1381 "cases49"
{
				 yyval.listptr = createlistnum
					(
					 yypvt[-0].realval
					);
				} break;
case 106:
# line 1388 "cases49"
{
				 yyval.listptr = createlistnum
					(
					 -yypvt[-0].realval
					);
				} break;
case 107:
# line 1395 "cases49"
{
				 yyval.listptr = createliststring
					(
					 yypvt[-0].strval
					);
				} break;
case 108:
# line 1402 "cases49"
{
				 yyval.listptr = createlistlist
					(
					 yypvt[-0].listptr
					);
				} break;
case 109:
# line 1411 "cases49"
{
				 yyval.repptr = yypvt[-0].repptr;
				} break;
case 110:
# line 1415 "cases49"
{
				 yyval.repptr = NULL;
				} break;
	}
	goto yystack;		/* reset registers in driver code */
}
