/*------------------------------------------------------*/
/*							*/
/*		EVALUATE.C				*/
/*							*/
/* The function evaluate is a function to evaluate an	*/
/* arithmetical expression stored in a tree. It returns	*/
/* the value as a float.				*/
/*------------------------------------------------------*/

float evaluate(tree,thecase)
struct arithnode *tree;
struct casenode *thecase;
{

 if (tree->contentstype == NUMVALUE)
   {
    return tree->numcontents;
   }
 else if (tree->contentstype == FIELDNAME)
   {
    return fieldvalue(tree->fieldname,thecase);
   }
 else if (tree->contentstype == PLUS)
   {
    return (evaluate(tree->left,thecase) + evaluate(tree->right,thecase));
   }
 else if (tree->contentstype == MINUS)
   {
    return (evaluate(tree->left,thecase) - evaluate(tree->right,thecase));
   }
 else if (tree->contentstype == MULTIPLY)
   {
    return (evaluate(tree->left,thecase) * evaluate(tree->right,thecase));
   }
 else if (tree->contentstype == DIVIDE)
   {
    if (evaluate(tree->right,thecase) == 0)
      {
       fprintf(stderr,"\n\nError - division by zero in expression\n");
       exit(1);
      }
    return (evaluate(tree->left,thecase) / evaluate(tree->right,thecase));
   }
 else if (tree->contentstype == UMINUS)
   {
    return (evaluate(tree->left,thecase) * -1.0);
   }
 else if (tree->contentstype == POWER)
   {
    return pow(evaluate(tree->left,thecase), evaluate(tree->right,thecase));
   }
 else
   {
    fprintf(stderr,"Huh? Unknown arithmetic operation type!\n");
    exit(1);
   }
}


/* find the numerical value of the named field in the solution fields of the given case */

float fieldvalue(fieldname,thecase)
char *fieldname;
struct casenode *thecase;
{
struct fieldnode *pointer;

 pointer = searchfield(fieldname,thecase->solnhead);

 if (pointer == NULL)
   {
    pointer = searchfield(fieldname,thecase->fieldhead);
    if (pointer == NULL)
      {
       fprintf(stderr,"Huh? - fieldname %s in arithmetical expression does not exist\n",
                      fieldname);
       exit(1);
      }
   }

 if (pointer->fieldtype == NUMTYPE)
   {
    return pointer->numcontents;
   }
 else
   {
    fprintf(stderr,"Error - Type mismatch - field %s in arithmetical expression\n",fieldname);
    exit(1);
   }
}


/* checkarith checks that all the fieldnames in an expression exist */

int checkarith(tree,thecase)	/* returns TRUE if OK, FALSE otherwise */
struct arithnode *tree;
struct casenode *thecase;
{

 if (tree->contentstype == NUMVALUE)
   {
    return TRUE;
   }
 else if (tree->contentstype == FIELDNAME)
   {
    if (exists(tree->fieldname,thecase->solnhead) == TRUE ||
        exists(tree->fieldname,thecase->fieldhead) == TRUE)
      {
       return TRUE;	/* field is in the case */
      }
    else
      {
       return FALSE;
      }
   }
 else if (tree->contentstype == PLUS ||
          tree->contentstype == MINUS ||
          tree->contentstype == MULTIPLY ||
          tree->contentstype == DIVIDE ||
          tree->contentstype == POWER)
   {
    if (checkarith(tree->left,thecase) == FALSE)
      {
       return FALSE;
      }
    return checkarith(tree->right,thecase);
   }
 else if (tree->contentstype == UMINUS)
   {
    return checkarith(tree->left,thecase);
   }
 else
   { 
    fprintf(stderr,"Huh? Unknown arithmetic operation type!\n");
    exit(1);
   }
}
