/*----------------------------------------------*/
/*						*/
/*		RULE REPAIR			*/
/*						*/
/* Version 1					*/
/*    This deals with the general repair rules.	*/
/* The general repair rules are applied to the	*/
/* problem fields and the solution fields. Those*/
/* fields for which there are no repair rules	*/
/* will not be changed.				*/
/*						*/
/* Version 2					*/
/*    The repairs are now done FROM the		*/
/* problem fields of the user's case TO the	*/
/* solution fields of the retrieved case. Also	*/
/* now a repair can cause a field to be created.*/
/*						*/
/* Version 3					*/
/*    The code has been re-written for the new	*/
/* format rules as discussed in thought2.	*/
/*						*/
/* Version 4					*/
/*    Abstract symbols can be used to match any	*/
/* of their specialisations.			*/
/*						*/
/* PLEASE NOTE:					*/
/*						*/
/* The way the rules work is now different.	*/
/* First, the NEWVALUE mechanism MUST be used	*/
/* to compare values with the user's fields.	*/
/* Otherwise the repair rules will look for	*/
/* combinations of values in the solution fields*/
/* of the reair case and			*/
/* NOT in the problem fields of the user's case.*/
/* Again, the NEWVALUE mechanism must be used	*/
/* to change the value to those in the user's	*/
/* problem fields.				*/
/*						*/
/* Version 5					*/
/*    The NEW VALUE mechanism has been replaced.*/
/* Now the problem fields of the user case are	*/
/* copied across to the problem fields of the	*/
/* repair case, and the comblist may refer to	*/
/* the problem fields. Arithmetical expressions	*/
/* may refer to problem fields but only the	*/
/* solution fields become changed.		*/
/*						*/
/* Modified for cases26				*/
/*						*/
/* Modified for cases32				*/
/*	This version implements local repair	*/
/* rules which are applied after the normal	*/
/* ones.					*/
/* Note that the repaired case does not have	*/
/* the local repair rules copied to it.		*/
/*						*/
/* Modified for cases34				*/
/*	This version abolishes th solution	*/
/* definition. Hence there is not much by way	*/
/* of type checking in repair rules.		*/
/*						*/
/* Modified for cases36				*/
/*	This gets rid of some compilation	*/
/* warning messages and abolishes the result	*/
/* value					*/
/*						*/
/* Modified for cases37				*/
/*	This changes the pr parameter to a list	*/
/* to be pretty-printed				*/
/*						*/
/* Modified for cases38				*/
/*	This implements hd and tl.		*/
/*						*/
/* Modified for cases39				*/
/*	This implements reselect.		*/
/*						*/
/* Modified for cases40				*/
/*	This implements local field definitions	*/
/*						*/
/* Modified for cases41				*/
/*	Abstractions are not associated with	*/
/* fields					*/
/*						*/
/* Modified for cases43				*/
/*	Implemented the extract operation.	*/
/*						*/
/* Modified for cases44				*/
/*	deletes symbol from list after extract	*/
/*						*/
/* Modified for cases45				*/
/*	New routine: expandlist()		*/
/*						*/
/* Modified for cases46				*/
/*	implements logical operations for	*/
/* combination nodes.				*/
/*						*/
/* Modified for cases47				*/
/*	if reselect fails then repairs are	*/
/* abandoned, rather than crashing out.		*/
/*						*/
/* Modified for cases48				*/
/*	enterfield is called differently.	*/
/*						*/
/* Modified for cases49				*/
/*	preprocess rules replace weight rules.	*/
/*----------------------------------------------*/

void rulerepair()	/* top level */
{
struct fieldnode *localfields,*lastfield;

/* if a reselect action is invoked then longjmp is invoked
   to jump back to here */

 setjmp(environment);

 reselectcase = FALSE;

 repcase = assemblerepcase(foundcase->name,
                        usercase->fieldhead,foundcase);
				/* copy retrieved case to repair case */

/* deal with local field definitions */

 if (foundcase->fielddefhead != NULL)
   {
    printf("\n\nPlease enter additional information...\n\n");

    /* prompt user to enter values for local fields */

    enterfield(foundcase->fielddefhead,&localfields);

    /* tag new local fields onto end of problem field list */

    lastfield = repcase->fieldhead;

    while (lastfield->fieldlink != NULL)
      {
       lastfield = lastfield->fieldlink;
      }

    lastfield->fieldlink = localfields;
   }


 printf("\n\n\nPERFORMING REPAIRS ON %s...\n\n\n",foundcase->name);

 if (reproot != NULL)
   {
    repairfields(reproot);		/* apply the default rules first */
   }

 if (foundcase != NULL && foundcase->rephead != NULL)
   {
    repairfields(foundcase->rephead);	/* apply the local rules */
   }

 fp = stdout;
 displaycase(repcase);
}


/* assemble the repair case, transfer any anonymous fields
   in foundcase, which are not defined in the usercase
   to the solution part of the repair case. */

struct casenode *assemblerepcase(name,probfields,foundcase)
buftype *name;
struct fieldnode *probfields;
struct casenode *foundcase;
{
struct casenode *pointer;
struct fieldnode *fieldptr,*fieldcopy;

 /* create repair case */

 pointer = assemblecase(name,probfields,foundcase->solnhead);

 /* if there are any anonymous fields in the problem part of the
    retrieved case which are not in the repair case, copy them
    to the solution part of the repair case.

 */

 fieldptr = foundcase->fieldhead;

 while (fieldptr != NULL)
   {
    if (anonymous(fieldptr->fieldname,foundcase))
      {
       if
         (
          (searchfield(fieldptr->fieldname,pointer->solnhead) == NULL) &&
          (searchfield(fieldptr->fieldname,pointer->fieldhead) == NULL)
         )
         {
          fieldcopy = copyfield(fieldptr);
          fieldcopy->fieldlink = pointer->solnhead;
          pointer->solnhead = fieldcopy;
         }
       if
         (
          (searchfield(fieldptr->fieldname,pointer->fieldhead) != NULL) &&
          (searchfield(fieldptr->fieldname,pointer->solnhead) != NULL)
         )
         {
          deletefield(fieldptr->fieldname,pointer->solnhead);
         }
      }
    fieldptr = fieldptr->fieldlink;
   }

 /* if there are any anonymous fields in the problem part of the
    repair case which are also in the problem part, delete them
    from the solution part */

 fieldptr = pointer->fieldhead;

 while (fieldptr != NULL)
   {
    if
      (
       anonymous(fieldptr->fieldname,pointer) &&
       (searchfield(fieldptr->fieldname,pointer->solnhead) != NULL)
      )
      {
       deletefield(fieldptr->fieldname,&(pointer->solnhead));
      }
    fieldptr = fieldptr->fieldlink;
   }


 return pointer;
}


/* delete a field from the list */

void deletefield(fieldname,fieldlist)
buftype *fieldname;
struct fieldnode **fieldlist;
{
struct fieldnode **pointer,*next;

 pointer = fieldlist;

 while (*pointer != NULL)
   {
    if (strcmp(fieldname,(*pointer)->fieldname) == 0)
      {
       break;
      }
    pointer = &((*pointer)->fieldlink);
   }

 if (*pointer == NULL)
   {
    fprintf(stderr,"Huh? - fieldname %s not found in deletefield()\n",
               fieldname);
    exit(1);
   }

 if ((*pointer)->fieldtype == STRTYPE ||
     (*pointer)->fieldtype == ENUMTYPE)
   {
    free((*pointer)->strcontents);	/* erase string node */
   }
 else if ((*pointer)->fieldtype == LISTTYPE)
   {
    deletelist((*pointer)->listcontents);
   }
 next = (*pointer)->fieldlink;
 free(*pointer);
 *pointer = next;
}


/* apply the pre-processing rules to usercase */

void preprocess()
{
struct fieldnode *fieldptr;

 if (preprocessroot != NULL)
   {
    repcase = usercase;	/* the repair case is the usercase */
    printf("\n\n\nPRE-PROCESSING usercase...\n\n\n");
    repairfields(preprocessroot);
 
		/* move the solution fields to the problem fields list */

    fieldptr = usercase->fieldhead;

    while (fieldptr->fieldlink != NULL)
      {
       fieldptr = fieldptr->fieldlink;
      }
    fieldptr->fieldlink = usercase->solnhead;
    usercase->solnhead = NULL;

		/* reset the repcase value */

    repcase = NULL;
   }
}


/* look for repair rules for all the fields */

void repairfields(reprules)
struct repnode *reprules;	/* pointer to the repair rules */
{
struct repnode *pointer;
int changemade;		/* set to true when a rule has been executed */
struct rulenamenode *nameptr;	/* pointer to list of rulenames */
int iterations;		/* count for iterations of outer repair loop */

 iterations = 0;

 do
   {
    nameptr = NULL;
    restartrepair = FALSE;

    do
      {
       changemade = FALSE;
       pointer = reprules;	/* for each repair rule */

       while (pointer != NULL)
         {
          if (inrulenamelist(pointer->rulename,nameptr) == FALSE)
            {
             if (examinereprule(pointer) == TRUE)
               {
                if (restartrepair == TRUE)
                  {
                   goto dorestart;
                  }
                if (reselectcase == TRUE)
                  {
                   deleterulenamelist(nameptr);
                   doreselectcase();
                   if (foundcase == NULL)
                     {
                      printf("Abandoning repairs!\n\n");
                      return;
                     }
                   longjmp(environment,1);
			/* this restarts the repairs */
                  }
                   
                changemade = TRUE;
                addrulename(pointer->rulename, &nameptr);
               }
            }
          pointer = pointer->replink;
         }
      }
    while (changemade == TRUE);	/* cycle until rules stop firing */

dorestart:

    deleterulenamelist(nameptr);
    iterations++;
   }
 while (restartrepair == TRUE && iterations <= MAXITER);

}


/* do a partial cleanup, reselect a case for repair */

void doreselectcase()
{
 /* remove the old foundcase from the list of case pointers */

 removecaseptr(foundcase);

 if (caseptrroot == NULL)
   {
    printf("\n\nreselect failed (sorry)\n");
    foundcase = NULL;
    return;
   }

 foundcase = matchcase(usercase);

 /* do a partial cleanup */

 deletecase(repcase);
 repcase = NULL;

 printf("\n\nRESELECTING CASE %s...\n\n",foundcase->name);
 displaycase(foundcase);
}


/* remove a case-pointer node from a list of case pointers */

void removecaseptr(caseptr)
struct casenode *caseptr;
{
struct caseptrnode **ptrptr,*next;

 ptrptr = &caseptrroot;

 while ((*ptrptr)->caseind != caseptr)
   {
    ptrptr = &((*ptrptr)->caseptrlink);
   }

 next = (*ptrptr)->caseptrlink;
 free(*ptrptr);
 *ptrptr = next;
}


/* look for a reprule to match this field and examine it */
/* return TRUE if a repair rule has fired */

int examinereprule(repptr)
struct repnode *repptr;
{

 if (matchcombtree(repptr->combtreeroot) == TRUE)
	/* if the list matches the user's values */
   {
    firerule(repptr);		/* do the repair */

    return TRUE;		/* signifies that a change has been made */
   }
 else
   {
    return FALSE;		/* no change made */
   }
}


/* execute the condition expression */

int matchcombtree(pointer)
struct combtreenode *pointer;
{
 if (pointer->optype == ITEM)
   {
    return matchcombitem(pointer->left);
   }
 else if (pointer->optype == NOT_OP)
   {
    if (matchcombtree(pointer->left) == TRUE)
      {
       return FALSE;
      }
    else
      {
       return TRUE;
      }
   }
 else if (pointer->optype == AND_OP)
   {
    if (matchcombtree(pointer->left) == FALSE)
      {
       return FALSE;
      }
    else
      {
       return matchcombtree(pointer->right);
      }
   }
 else if (pointer->optype == OR_OP)
   {
    if (matchcombtree(pointer->left) == TRUE)
      {
       return TRUE;
      }
    else
      {
       return matchcombtree(pointer->right);
      }
   }
 else
   {
    fprintf(stderr,"Huh? - Unknown logical operator type in matchcombtree()\n");
   }
}


/* match a field value with a combnode */
/* return true if the match succeeds */

int matchcombitem(combptr)
struct combnode *combptr;
{
 if (combptr->undefined == TRUE)
   {
    return matchcombundefined(combptr);	/* TRUE if symbol is undefined */
   }
 else if (combptr->listop == TRUE)
   {
    return matchlistop(combptr);	/* TRUE if list operation matches
						enumeration value */
   }
 else if (combptr->fieldtype == ENUMTYPE)
   {
    return matchcombenum(combptr->fieldname,combptr->strcontents);
					/* TRUE if enumeration values match */
   }
 else if (combptr->fieldtype == NUMTYPE)
   {
    return matchcombnum(combptr);	/* TRUE if comparison evaluates to TRUE */
   }
 else if (combptr->fieldtype == STRTYPE)
   {
    return matchcombstr(combptr);	/* TRUE if string values match */
   }
 else if (combptr->fieldtype == LISTTYPE)
   {
    return matchcomblist(combptr);	/* TRUE if lists match */
   }
 else
   {
    fprintf(stderr,"Huh? - Unknown type in combnode\n");
    exit(1);
   }
}


/* return TRUE if fieldname is undefined */

int matchcombundefined(combptr)
struct combnode *combptr;
{

 if (searchfield(combptr->fieldname,repcase->solnhead) != NULL ||
     searchfield(combptr->fieldname,repcase->fieldhead) != NULL)
   {
    return FALSE;	/* field exists */
   }
 else
   {
    return TRUE;
   }
}


/* invoke a list operation for the given listop list */
/* returns a pointer to a pointer to the result */
/* (or a pointer to a float if the resulttype is NUMTYPE) */
/* the pointer must be typecast according to the type */
/* returned by resulttype */


char *dolistop(listopptr,resulttype)
struct listopnode *listopptr;
int *resulttype;
{
struct fieldnode *pointer;
struct listnode **listptrptr;
int tailtype;

 if (listopptr->optype == FIELDNAME)
   {
    pointer = searchfield(listopptr->fieldname,repcase->solnhead);

    if (pointer == NULL)
      {
       pointer = searchfield(listopptr->fieldname,repcase->fieldhead);
      }

    if (pointer == NULL)
      {
       return NULL;
      }
    else if (pointer->fieldtype != LISTTYPE)
      {
       fprintf(stderr,"\nError - field %s not list type\n",pointer->fieldname);
       exit(1);
      }
    else
      {
       *resulttype = LISTTYPE;
       return (char *) &(pointer->listcontents);
			/* returns pointer to pointer to list */
      }
   }

 else if (listopptr->optype == HEAD)
   {
    listptrptr = (struct listnode **) dolistop(listopptr->listoplink,&tailtype);
    if (listptrptr == NULL)
      {
       return NULL;
      }
    else if (tailtype != LISTTYPE)
      {
       fprintf(stderr,"\nError - hd operation performed on non-list type\n");
       exit(1);
      }
    else
      {
       *resulttype = (*listptrptr)->listtype;
       if (*resulttype == ENUMTYPE || *resulttype == STRTYPE)
         {
          return (char *) &((*listptrptr)->strcontents);
         }
       else if (*resulttype == LISTTYPE)
         {
          return (char *) &((*listptrptr)->listitem);
         }
       else
         {
          return (char *) &((*listptrptr)->numcontents);
         }
      }
   }

 else if (listopptr->optype == TAIL)
   {
    listptrptr = (struct listnode **) dolistop(listopptr->listoplink,&tailtype);
    if (listptrptr == NULL)
      {
       return NULL;
      }
    else if (tailtype != LISTTYPE)
      {
       fprintf(stderr,"\nError - tl operation performed on non-list type\n");
       exit(1);
      }
    else
      {
       *resulttype = LISTTYPE;
       return (char *) &((*listptrptr)->listlink);
      }
   }
}


/* return TRUE if list operation matches enumeration value */

int matchlistop(combptr)
struct combnode *combptr;
{
int resulttype;
char *pointer;

 pointer = dolistop(combptr->listophead,&resulttype);

 if (pointer == NULL)
   {
    return FALSE;
   }

 if (resulttype != combptr->fieldtype)
   {
    return FALSE;
   }

 if (resulttype == NUMTYPE)
   {
    return matchlistopexp(*((float *) pointer), combptr);
   }
 else if (resulttype == ENUMTYPE)
   {
    if (strcmp(* ((char **) pointer),combptr->strcontents) == 0)
      {
       return TRUE;
      }
    else if (abstract(combptr->strcontents) == TRUE)
      {
       return matchspecialisations(combptr->strcontents,* ((char **) pointer));
      }
    else
      {
       return FALSE;
      }
   }
 else if (resulttype == STRTYPE)
   {
    if (strcmp(* ((char **) pointer),combptr->strcontents) == 0)
      {
       return TRUE;
      }
    else
      {
       return FALSE;
      }
   }
 else
   {
    return matchlist(*((struct listnode **) pointer),combptr->listvalue);
   }
}


/* return TRUE if listop value compares with expression */

int matchlistopexp(value,combptr)
float value;
struct combnode *combptr;
{

 if (checkarith(combptr->right,repcase) == FALSE)
   {
    return FALSE;
   }

 if (combptr->operator == '=')
   {
    if (value == evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == NOTEQUAL)
   {
    if (value != evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == '<')
   {
    if (value < evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == LESSOREQUAL)
   {
    if (value <= evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == '>')
   {
    if (value > evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == MOREOREQUAL)
   {
    if (value >= evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else
   {
    fprintf(stderr,"\n\nHuh? - combnode contains unknown operator\n\n");
    exit(1);
   }
}


/* return TRUE if enumeration types match */

int matchcombenum(fieldname,value)
buftype *fieldname,*value;
{
struct fieldnode *pointer;

 pointer = searchfield(fieldname, repcase->solnhead);
 if (pointer == NULL)
   {
    pointer = searchfield(fieldname, repcase->fieldhead);
   }

 if (pointer == NULL)	/* field not defined in solution fields */
   {
    return FALSE;
   }
 else
   {
    if (strcmp(pointer->strcontents, value) == 0)
      {
       return TRUE;	/* enumeration strings match */
      }
    else
      {
       if (abstract(value) == TRUE)
         {
          return matchspecialisations(value,pointer->strcontents);
         }
       else
         {
          return FALSE;
         }
      }
   }
}


/* match an enumeration against the specialisations of an abstract symbol */

int matchspecialisations(absvalue,evalue)
buftype *absvalue,*evalue;
{
struct modnode *modptr;
struct enumnode *enumptr;

 modptr = modroot;

 while (modptr != NULL)
   {
    if (strcmp(modptr->abstraction, absvalue) == 0)
      {
       break;
      }
    modptr = modptr->modlink;
   }

 if (modptr == NULL)
   {
    fprintf(stderr,"Huh? Abstraction %s not found in modification rules\n",absvalue);
    exit(1);
   }

 enumptr = modptr->enumhead;

 while (enumptr != NULL)
   {
    if (strcmp(enumptr->evalue, evalue) == 0)
      {
       return TRUE;
      }
    else
      {
       if (abstract(enumptr->evalue) == TRUE)
         {
          if (matchspecialisations(enumptr->evalue,evalue) == TRUE)
            {
             return TRUE;
            }
         }
      }
    enumptr = enumptr->enumlink;
   }
 return FALSE;
}


/* return TRUE if number type is within range */

int matchcombnum(combptr)
struct combnode *combptr;
{

 if (checkarith(combptr->left,repcase) == FALSE ||	/* both sides can be */
     checkarith(combptr->right,repcase) == FALSE)	/* evalauted.	*/
   {
    return FALSE;
   }

 if (combptr->operator == '=')
   {
    if (evaluate(combptr->left,repcase) == evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == NOTEQUAL)
   {
    if (evaluate(combptr->left,repcase) != evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == '<')
   {
    if (evaluate(combptr->left,repcase) < evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == LESSOREQUAL)
   {
    if (evaluate(combptr->left,repcase) <= evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == '>')
   {
    if (evaluate(combptr->left,repcase) > evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else if (combptr->operator == MOREOREQUAL)
   {
    if (evaluate(combptr->left,repcase) >= evaluate(combptr->right,repcase))
       return TRUE;
    else
       return FALSE;
   }
 else
   {
    fprintf(stderr,"\n\nHuh? - combnode contains unknown operator\n\n");
    exit(1);
   }
}


/* return TRUE if strings match */

int matchcombstr(combptr)
struct combnode *combptr;
{
struct fieldnode *pointer;

 pointer = searchfield(combptr->fieldname, repcase->solnhead);
 if (pointer == NULL)
   {
    pointer = searchfield(combptr->fieldname, repcase->fieldhead);
   }

 if (pointer == NULL)	/* field not defined in solution list */
   {
    return FALSE;
   }
 else
   {
    if (strcmp(pointer->strcontents, combptr->strcontents) == 0)
      {
       return TRUE;
      }
    else
      {
       return FALSE;
      }
   }
}


/* return TRUE if lists match */

int matchcomblist(combptr)
struct combnode *combptr;
{
struct fieldnode *pointer;

 pointer = searchfield(combptr->fieldname, repcase->solnhead);
 if (pointer == NULL)
   {
    pointer = searchfield(combptr->fieldname, repcase->fieldhead);
   }

 if (pointer == NULL)	/* field not defined in solution list */
   {
    return FALSE;
   }
 else
   {
    return matchlist(pointer->listcontents, combptr->listvalue);
   }
}


/* perform the repair */

void firerule(repptr)
struct repnode *repptr;
{
 firereplist(repptr->replisthead);
}


/* print a list with list expansion on. */

void expandlist(list,thecase)
struct listnode *list;
struct casenode *thecase;
{
int temp;

    temp = listexpansion;
    listexpansion = TRUE;
    displaylist(list,thecase);
    listexpansion = temp;
}


/* perform the repairs in the repair list */

void firereplist(replistptr)
struct replistnode *replistptr;
{
char *pointer;
int resulttype;

 if (replistptr == NULL)
   {
    return;
   }
 else if (replistptr->listop == TRUE)
   {
    pointer = dolistop(replistptr->listophead,&resulttype);
    if (pointer == NULL)
      {
       printf("\npointer is NULL\n");	/* do nothing */
      }
    else if (resulttype != replistptr->fieldtype)
      {
       fprintf(stderr,"\nError - type mismatch changing list element\n");
       exit(1);
      }
    else if (resulttype == STRTYPE || resulttype == ENUMTYPE)
      {
       strcpy(* ((char **) pointer),replistptr->strvalue);
      }
    else if (resulttype == LISTTYPE)
      {
       deletelist(* ((struct listnode **) pointer));
       *((struct listnode **) pointer) = copylist(replistptr->tolist);
      }
    else
      {
       if (checkarith(replistptr->exptree,repcase) == TRUE)
         {
          *((float *) pointer) = evaluate(replistptr->exptree,repcase);
         }
      }
   }
 else if (replistptr->restart == TRUE)		/* restart the repair rules */
   {
    restartrepair = TRUE;
    return;
   }
 else if (replistptr->reselect == TRUE)		/* reselect case */
   {
    reselectcase = TRUE;
    return;
   }
 else if (replistptr->print == TRUE)		/* execute print command */
   {
    fp = stdout;
    expandlist(replistptr->tolist,repcase);
   }
 else if (replistptr->undefined == TRUE)	/* field becomes undefined */
   {
    deletenode(replistptr->fieldname,&(repcase->solnhead));
   }
 else if (replistptr->extract == TRUE)		/* extract symbol from list */
   {
    extractenum(replistptr->fieldname,
                replistptr->abstraction,
                replistptr->strvalue);
   }
 else if (replistptr->weightchange == TRUE)	/* change weight of a field */
   {
    changeweight(replistptr->fieldname,replistptr->weight);
   }
 else if (replistptr->fieldtype == STRTYPE)
   {
    changetostring(replistptr->fieldname,replistptr->strvalue);/* new string value */
   }
 else if (replistptr->fieldtype == ENUMTYPE)
   {
    changetoenum(replistptr->fieldname,replistptr->strvalue);	/* new enumeration value */
   }
 else if (replistptr->fieldtype == NUMTYPE)
   {
    changetonum(replistptr->fieldname,replistptr->exptree);	/* new number value */
   }
 else if (replistptr->fieldtype == LISTTYPE)
   {
    changetolist(replistptr->fieldname,replistptr->tolist);	/* new list value */
   }
 else
   {
    fprintf(stderr,"Huh? - Unknown replist type\n");
    exit(1);
   }

 firereplist(replistptr->replistlink);
}


/* change the weight of a field (cases49) */

void changeweight(fieldname,weight)
buftype *fieldname;
float weight;
{
struct typenode *thefield;

 thefield = findfield(fieldname,probdefnroot);

 if (thefield == NULL)
   {
    fprintf(stderr,"Huh? - field %s not defined in changefield()\n\n",
            fieldname);
    exit(1);
   }
 else
   {
    thefield->weight = weight;
    printf("Changed weight of %s to %.2f\n",fieldname,weight);
   }
}


/* change a string value in the repair case */

void changetostring(fieldname,strvalue)
buftype *fieldname;
buftype *strvalue;
{
struct fieldnode *pointer;

 if (findfield(fieldname,probdefnroot) != NULL ||
     findfield(fieldname,foundcase->fielddefhead) != NULL)
				/* field is problem field */
   {
    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer == NULL)
      {
       pointer = createnewfieldnode(fieldname,STRTYPE,&(repcase->fieldhead));
       pointer->strcontents = createbuftypenode();
      }
   }
 else
   {

    /* if the field exists as an anyonymous field, delete it */

    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer != NULL)
      {
       deletenode(fieldname, &(repcase->fieldhead));
      }

    pointer = searchfield(fieldname, repcase->solnhead);
    if (pointer != NULL)
      {
       if (pointer->fieldtype != STRTYPE)
         {
          deletenode(fieldname, &(repcase->solnhead));
          pointer = NULL;
         }
      }
    if (pointer == NULL)	/* field not in solution list */
      {
       pointer = createnewfieldnode(fieldname,STRTYPE,&(repcase->solnhead));
				/* so create a field */
       pointer->strcontents = createbuftypenode();
      }
   }

 pointer->fieldtype = STRTYPE;
 strcpy(pointer->strcontents, strvalue);
}


/* change an enumeration value in the repair case */

void changetoenum(fieldname, enumvalue)
buftype *fieldname;
buftype *enumvalue;
{
struct fieldnode *pointer;

 if (findfield(fieldname,probdefnroot) != NULL ||
     (foundcase != NULL &&
      findfield(fieldname,foundcase->fielddefhead) != NULL))
				/* field is problem field */
   {
    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer == NULL)
      {
       pointer = createnewfieldnode(fieldname,ENUMTYPE,&(repcase->fieldhead));
       pointer->strcontents = createbuftypenode();
      }
   }
 else
   {

    /* if the field exists as an anyonymous field, delete it */

    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer != NULL)
      {
       deletenode(fieldname, &(repcase->fieldhead));
      }

    pointer = searchfield(fieldname, repcase->solnhead);
    if (pointer != NULL)
      {
       if (pointer->fieldtype != ENUMTYPE)
         {
          deletenode(fieldname, &(repcase->solnhead));
          pointer = NULL;
         }
      }
    if (pointer == NULL)        /* field not in solution list */
      {
       pointer = createnewfieldnode(fieldname,ENUMTYPE,&(repcase->solnhead));
                                /* so create a field */
       pointer->strcontents = createbuftypenode();
      }
   }

 pointer->fieldtype = ENUMTYPE;
 strcpy(pointer->strcontents, enumvalue);
}


/* perform an extraction to get a enumeration value from a list */

void extractenum(fieldname,abstraction,listname)
buftype *fieldname,*abstraction,*listname;
{
struct fieldnode *listfield;
struct listnode *listptr;

	/* find the list containing the symbols */

 listfield = searchfield(listname, repcase->fieldhead);
 if (listfield == NULL)
   {
    listfield = searchfield(listname, repcase->solnhead);
    if (listfield == NULL)
      {
       fprintf(stderr,"Error - list field %s not found doing extract\n",listname);
       exit(1);
      }
   }

 if (listfield->fieldtype != LISTTYPE)
   {
    fprintf(stderr,"Error - %s should be list type doing extract\n",listname);
    exit(1);
   }
	/* search through list looking for symbol */

 listptr = listfield->listcontents;

 while (listptr != NULL)
   {
    if (listptr->listtype == ENUMTYPE &&
        matchspecialisations(abstraction,listptr->strcontents) == TRUE)
      {
       break;
      }
    listptr = listptr->listlink;
   }

	/* reached end of list without finding anything appropriate */

 if (listptr == NULL)
   {
    return;
   }

	/* check that enumeration symbol is OK for this field */

 if (findfield(fieldname,repcase->fielddefhead) ||
     findfield(fieldname,probdefnroot))
   {
    localcheckenum(fieldname,listptr->strcontents,repcase);
   }

	/* change the value of the field */

 changetoenum(fieldname,listptr->strcontents);

	/* delete the node from the list */

 deletelistitem(listptr,&(listfield->listcontents));
}


/* delete an item from a list */

void deletelistitem(item,listhead)
struct listnode *item,**listhead;
{
struct listnode **listptrptr,*temp;

	/* delete the contents of the list */

 if (item->listtype == STRTYPE ||
     item->listtype == ENUMTYPE)
   {
    free(item->strcontents);
   }
 else if (item->listtype == LISTTYPE)
   {
    deletelist(item->listitem);
   }

	/* find the item within the list */

 listptrptr = listhead;

 while (*listptrptr != NULL)
   {
    if (*listptrptr == item)
      {
       break;
      }
    listptrptr = &((*listptrptr)->listlink);
   }

 if (*listptrptr == NULL)
   {
    fprintf(stderr,"Huh? - item not found in deletelistitem()\n");
    exit(1);
   }
 else	/* delete the item */
   {
    temp = item->listlink;
    free(item);
    *listptrptr = temp;
   }
}


/* change a number value in the repair case */

void changetonum(fieldname, exptree)
buftype *fieldname;
struct arithnode *exptree;
{
struct fieldnode *pointer;

 if (findfield(fieldname,probdefnroot) != NULL ||
     findfield(fieldname,foundcase->fielddefhead) != NULL)
				/* field is problem field */
   {
    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer == NULL) 
      {
       pointer = createnewfieldnode(fieldname,NUMTYPE,&(repcase->fieldhead));
      }
   }
 else
   {

    /* if the field exists as an anyonymous field, delete it */

    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer != NULL)
      {
       deletenode(fieldname, &(repcase->fieldhead));
      }

    pointer = searchfield(fieldname, repcase->solnhead);
    if (pointer != NULL)
      {
       if (pointer->fieldtype != NUMTYPE)
         {
          deletenode(fieldname, &(repcase->solnhead));
          pointer = NULL;
         }
      }
    if (pointer == NULL)        /* field not in solution list */
      {
       pointer = createnewfieldnode(fieldname,NUMTYPE,&(repcase->solnhead));
                                /* so create a field */
      }
   }

 if (checkarith(exptree,repcase) == TRUE)
   {
    pointer->numcontents = evaluate(exptree,repcase);
    pointer->fieldtype = NUMTYPE;
   }
 else
   {
    deletenode(fieldname, &(repcase->solnhead));
   }
}


/* change a list value in the repair case */

void changetolist(fieldname, listptr)
buftype *fieldname;
struct listnode *listptr;
{
struct fieldnode *pointer;

 if (findfield(fieldname,probdefnroot) != NULL ||
     findfield(fieldname,foundcase->fielddefhead) != NULL)
				/* field is problem field */
   {
    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer == NULL)
      {
       pointer = createnewfieldnode(fieldname,LISTTYPE,&(repcase->fieldhead));
      }
    else
      {
       deletelist(pointer->listcontents);
      }
   }
 else
   {

    /* if the field exists as an anyonymous field, delete it */

    pointer = searchfield(fieldname, repcase->fieldhead);
    if (pointer != NULL)
      {
       deletenode(fieldname, &(repcase->fieldhead));
      }

    pointer = searchfield(fieldname, repcase->solnhead);
    if (pointer != NULL)
      {
       if (pointer->fieldtype != LISTTYPE)
         {
          deletenode(fieldname, &(repcase->solnhead));
          pointer = NULL;
         }
       else
         {
          deletelist(pointer->listcontents);
         }
      }
    if (pointer == NULL)        /* field not in solution list */
      {
       pointer = createnewfieldnode(fieldname,LISTTYPE,&(repcase->solnhead));
                                /* so create a field */
      }
   }

 pointer->fieldtype = LISTTYPE;
 pointer->listcontents = copylist(listptr);	/* make a copy of the list */
}


/* create a fieldnode in the repair case */

struct fieldnode *createnewfieldnode(fieldname,fieldtype,fieldptr)
buftype *fieldname;
int fieldtype;
struct fieldnode **fieldptr;
{
struct fieldnode *pointer;

 pointer = (struct fieldnode *) calloc(1, sizeof(struct fieldnode));
				/* create a fieldnode */

 pointer->fieldlink = *fieldptr;
 *fieldptr = pointer;
 strcpy(pointer->fieldname,fieldname);
 pointer->fieldtype = fieldtype;
 return pointer;
}


/* search for a field with a given name */

struct fieldnode *searchfield(fieldname,list)
char *fieldname;
struct fieldnode *list;
{
struct fieldnode *pointer;

 pointer = list;

 while (pointer != NULL)
   {
    if (strcmp(pointer->fieldname,fieldname) == 0)
       return pointer;
    pointer = pointer->fieldlink;
   }

 return NULL;
}


/* delete a node from a fieldlist */

void deletenode(fieldname,tolist)
char *fieldname;
struct fieldnode **tolist;
{
struct fieldnode *pointer,*field;

 field = searchfield(fieldname,*tolist);

 if (field == NULL)	/* field doesn't exist so do nothing */
    return;

 deletenodeinfo(field);	/* delete list/string buffer first */

 if (field == *tolist)	/* at top of list */
   {
    *tolist = field->fieldlink;
    free(field);
   }
 else
   {
    pointer = *tolist;
    while (pointer->fieldlink != field)
      {
       pointer = pointer->fieldlink;
      }
    pointer->fieldlink = field->fieldlink;
    free(field);
   }
}


/* delete the string buffer and list attached to fieldnode */

void deletenodeinfo(fieldptr)
struct fieldnode *fieldptr;
{
 if (fieldptr->fieldtype == STRTYPE ||
     fieldptr->fieldtype == ENUMTYPE)
   {
    free(fieldptr->strcontents);
   }
 else if (fieldptr->fieldtype == LISTTYPE)
   {
    deletelist(fieldptr->listcontents);
   }
 else
   {
    /* do nothing */
   }
}


/* delete the contents of a list */

void deletelist(listptr)
struct listnode *listptr;
{
 if (listptr == NULL)
   {
    return;	/* null list */
   }
 else if (listptr->listtype == STRTYPE ||
          listptr->listtype == ENUMTYPE)
   {
    free(listptr->strcontents);
    deletelist(listptr->listlink);
    free(listptr);
   }
 else if (listptr->listtype == LISTTYPE)
   {
    deletelist(listptr->listitem);
    deletelist(listptr->listlink);
    free(listptr);
   }
 else
   {
    deletelist(listptr->listlink);
    free(listptr);
   }
}


/* copy the contents of a list */
struct listnode *copylist(thelist)
struct listnode *thelist;
{
struct listnode *listptr,*pointer;
char *string;

 if (thelist == NULL)
   {
    return NULL;
   }
 else if (thelist->listtype == ENUMTYPE)
   {
    string = createstringnode();
    strcpy(string,thelist->strcontents);
    pointer = createlistenum(string, NULL);
    pointer->listlink = copylist(thelist->listlink);
    return pointer;
   }
 else if (thelist->listtype == NUMTYPE)
   {
    pointer = createlistnum(thelist->numcontents, NULL);
    pointer->listlink = copylist(thelist->listlink);
    return pointer;
   }
 else if (thelist->listtype == STRTYPE)
   {
    string = createstringnode();
    strcpy(string,thelist->strcontents);
    pointer = createliststring(string, NULL);
    pointer->listlink = copylist(thelist->listlink);
    return pointer;
   }
 else	/* list item is a list */
   {
    listptr = copylist(thelist->listitem);
    pointer = createlistlist(listptr, NULL);
    pointer->listlink = copylist(thelist->listlink);
    return pointer;
   }
}


/* make a copy of a case and return a pointer to it */
/* local repair rules are not copied */

struct casenode *copycase(thecase)
struct casenode *thecase;
{
struct casenode *pointer;

 pointer = createcasenode(thecase->name);

 pointer->fieldhead = copyfieldlist(thecase->fieldhead);
 pointer->solnhead = copyfieldlist(thecase->solnhead);
 pointer->fielddefhead = NULL;
 pointer->rephead = NULL;
 return pointer;
}


/* make a copy of a list of fields & return a pointer to it*/

struct fieldnode *copyfieldlist(thisfield)
struct fieldnode *thisfield;
{
struct fieldnode *pointer;

 if (thisfield == NULL)
   {
    return NULL;	/* no more nodes */
   }
 else
   {
    pointer = copyfield(thisfield);
    pointer->fieldlink = copyfieldlist(thisfield->fieldlink);
    return pointer;
   }
}


/* make a copy of a field & return a pointer to it*/

struct fieldnode *copyfield(thisfield)
struct fieldnode *thisfield;
{
struct fieldnode *pointer;
char *strptr;
struct listnode *listptr;

 if (thisfield == NULL)
   {
    return NULL;        /* null field */
   }

 if (thisfield->fieldtype == STRTYPE)
   {
    strptr = createstringnode();
    strcpy(strptr,thisfield->strcontents);
    pointer = createstrfield(thisfield->fieldname,strptr);
   }
 else if (thisfield->fieldtype == NUMTYPE)
   {
    pointer = createnumfield(thisfield->fieldname, thisfield->numcontents);
   }
 else if (thisfield->fieldtype == LISTTYPE)
   {
    listptr = copylist(thisfield->listcontents);
    pointer = createlistfield(thisfield->fieldname, listptr);
   }
 else if (thisfield->fieldtype == ENUMTYPE)   /* enumerated type */
   {
    strptr = createstringnode();
    strcpy(strptr,thisfield->strcontents);
    pointer = createenumfield(thisfield->fieldname, strptr);
   }
 else
   {
    fprintf(stderr,"\nHuh? - Unknown type doing copylist()\n");
    exit(1);
   }

 return pointer;
}


/* add a rulename to the head of a linked list of rulenames */

void addrulename(rulename,rulelist)
buftype *rulename;
struct rulenamenode **rulelist;
{
struct rulenamenode *pointer;

 pointer = (struct rulenamenode *) calloc(1, sizeof(struct rulenamenode));
			/* allocate space for node */

 strcpy(pointer->rulename,rulename);
 pointer->rulenamelink = *rulelist;
 *rulelist = pointer;
}


/* return true if rulename is in rulename list */

int inrulenamelist(rulename,rulelist)
buftype *rulename;
struct rulenamenode *rulelist;
{
 if (rulelist == NULL)
   {
    return FALSE;
   }
 else if (strcmp(rulename, rulelist->rulename) == 0)
   {
    return TRUE;
   }
 else
   {
    return inrulenamelist(rulename,rulelist->rulenamelink);
   }
}


/* delete the rulename list */

void deleterulenamelist(rulelist)
struct rulenamenode *rulelist;
{
 if (rulelist == NULL)
   {
    return;
   }
 else
   {
    deleterulenamelist(rulelist->rulenamelink);
    free(rulelist);
   }
}
